""""""

import pytest
import numpy as np

from pybryt import *
from pybryt import complexities as cplx
from pybryt.execution.complexity import TimeComplexityResult

from .utils import check_obj_attributes


def generate_complexity_footprint(name, t_transform, max_exp=8):
    footprint = []
    for i, e in enumerate(range(1, max_exp + 1)):
        n = 10 ** e
        t = t_transform(n)
        footprint.append((TimeComplexityResult(name, n, 0, t), i))
    return footprint


def test_complexity_abc():
    """
    """
    Annotation.reset_tracked_annotations()

    a = TimeComplexity(cplx.constant, name="foo")
    check_obj_attributes(a, {
        "name": "foo",
        "complexity": cplx.constant,
    })

    b = TimeComplexity(cplx.constant, name="foo")
    assert a == b

    b.complexity = cplx.linear
    assert a != b

    b.name = "bar"
    assert a != b

    b.complexity = cplx.constant
    assert a != b

    # test constructor errors
    with pytest.raises(ValueError, match="Complexity annotations require a 'name' kwarg"):
        TimeComplexity(cplx.constant)

    with pytest.raises(ValueError, match="Invalid valid for argument 'complexity': 1"):
        TimeComplexity(1, name="foo")


def test_time_complexity():
    Annotation.reset_tracked_annotations()

    a = TimeComplexity(cplx.constant, name="foo")

    footprint = generate_complexity_footprint("foo", lambda v: 1012)
    res = a.check(footprint)
    assert res.satisfied
    assert res.value == cplx.constant

    footprint.append((np.random.uniform(size=100), 9))
    footprint.append((TimeComplexityResult("bar", 10, 0, 10 ** 3), 10))
    res = a.check(footprint)
    assert res.satisfied
    assert res.value == cplx.constant

    footprint = generate_complexity_footprint("foo", np.log2)
    res = a.check(footprint)
    assert not res.satisfied
    assert res.value == cplx.logarithmic

    footprint = generate_complexity_footprint("foo", lambda v: v * np.log2(v))
    res = a.check(footprint)
    assert not res.satisfied
    assert res.value == cplx.linearithmic

    a.complexity = cplx.exponential
    res = a.check(footprint)
    assert not res.satisfied
    assert res.value == cplx.linearithmic


def test_alias():
    import pybryt.complexities as cplx2
    assert cplx.complexity_classes is cplx2.complexity_classes



"""Annotations for asserting the presence of a value"""

__all__ = ["Value", "Attribute"]

import dill
import pandas as pd
import numpy as np

from collections.abc import Iterable
from copy import copy
from typing import Any, Dict, List, Optional, Tuple, Union

from .annotation import Annotation, AnnotationResult
from .invariants import invariant


class Value(Annotation):
    """
    Annotation class for asserting that a value should be observed.

    Indicates that a value passed to the constructor should be observed while tracing through the
    students' code. Values can be of any type that is picklable by ``dill``. Values can specify a
    list of :ref:`invariants<invariants>` that will allow objects to be considered "equal." For 
    values that support arithemtic operators, absolute tolerances can be specified as well.

    Numeric tolerances are computed as with ``numpy.allcose``, where the value is considered "equal 
    enough" if it is within :math:`v \\pm (\\texttt{atol} + \\texttt{rtol} \\cdot |v|)`, where 
    :math:`v` is the value of the annotation.

    Args:
        value (``object``): the value that should be observed
        atol (``float`` or ``int``, optional): absolute tolerance for numeric values
        rtol (``float`` or ``int``, optional): relative tolerance for numeric values
        invariants (``list[invariant]``): invariants for 
            this value
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.annotations.annotation.Annotation>` constructor
    """

    intial_value: Any
    """a copy of the value passed to the constructor"""

    _values: List[Any]
    """
    a list of values that resulted from passing the intial value through the series of invariants 
    specified in the constructor
    """

    atol: Optional[Union[float, int]]
    """absolute tolerance for numeric values"""

    rtol: Optional[Union[float, int]]
    """relative tolerance for numeric values"""

    invariants: List[invariant]
    """the invariants for this value"""

    def __init__(
        self, value: Any, atol: Optional[Union[float, int]] = None, rtol: Optional[Union[float, int]] = None, 
        invariants: List[invariant] = [], **kwargs
    ):
        try:
            dill.dumps(value)
        except Exception as e:
            raise ValueError(f"Values must be serializable but the following error was thrown during serialization:\n{e}")

        self.initial_value = copy(value)
        self._values = [self.initial_value]
        self.atol = atol
        self.rtol = rtol
        self.invariants = invariants

        for inv in self.invariants:
            self._values = inv(self._values)

        super().__init__(**kwargs)

    @property
    def children(self) -> List[Annotation]:
        return []

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts this annotation's details to a JSON-friendly dictionary format.

        Output dictionary contains the annotation's name, group, limit number, success message, and
        failure message, as well as an ``invariants`` key with a list of the names of all invariants
        used in this value annotation and a ``atol`` key with this annotation's tolerance. This 
        dictionary does *not* contain the value being tracked.

        Returns:
            ``dict[str, object]``: the dictionary representation of this annotation
        """
        d = super().to_dict()
        d.update({
            "invariants": [inv.__name__ for inv in self.invariants],
            "atol": self.atol,
            "rtol": self.rtol,
        })
        return d

    def check(self, observed_values: List[Tuple[Any, int]]) -> AnnotationResult:
        """
        Checks that the value tracked by this annotation occurs in the list of observed values.

        Checks that the value under the conditions of its invariants occurs in the list of tuples of
        observed values and timestamps ``observed_values``. Creates and returns an 
        :py:class:`AnnotationResult<pybryt.AnnotationResult>` object with the results of this check.

        Args:
            observed_values (``list[tuple[object, int]]``): a list of tuples of values observed
                during execution and the timestamps of those values
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation based on ``observed_values``
        """
        satisfied = [self._check_observed_value(v) for v in observed_values]
        if not any(satisfied):
            return AnnotationResult(False, self)

        first_satisfier = satisfied.index(True)
        return AnnotationResult(True, self, observed_values[first_satisfier][0], observed_values[first_satisfier][1])
    
    def __eq__(self, other: Any) -> bool:
        """
        Checks whether this annotation is equal to another object.

        To be equal to a ``Value``, the other object must also be a ``Value`` object, have the same
        set of invariants, the same tolerance, and the same initial value.

        Args:
            other (``object``): the object to compare to

        Returns:
            ``bool``: whether the objects are equal
        """
        return super().__eq__(other) and self.invariants == other.invariants and \
            self.check_values_equal(self.initial_value, other.initial_value) and \
            self.atol == other.atol and self.rtol == other.rtol

    def _check_observed_value(self, observed_value: Tuple[Any, int]) -> bool:
        """
        Checks whether a single observed value tuple satisfies this value.

        Applies all invariants to the first element of the tuple and then checks whether any of the
        resulting values match and of the values in ``self._values``.

        Args:
            observed_value (``tuple[object, int]``): the observed value tuple

        Returns:
            ``bool``: whether the value matched
        """
        other_values = [observed_value[0]]
        for inv in self.invariants:
            other_values = inv(other_values)
        
        for value in self._values:
            for other_value in other_values:
                if self.check_values_equal(value, other_value, self.atol, self.rtol):
                    return True

        return False

    @staticmethod
    def check_values_equal(value, other_value, atol = None, rtol = None) -> bool:
        """
        Checks whether two objects are equal. 
        
        If the values are both numeric (numerics, arrays, etc.) and ``atol`` and/or ``rtol`` are 
        specified, the values are considered equal if ``other_value`` is within the tolerance
        bounds of ``value``.

        Args:
            value (``object``): the first object to compare
            other_value (``object``): the second object to compare
            atol (``float``, optional): the absolute tolerance for numeric values
            rtol (``float``, optional): the relative tolerance for numeric values
        """
        if isinstance(value, Iterable) ^ isinstance(other_value, Iterable):
            return False

        if atol is None:
            atol = 0
        if rtol is None:
            rtol = 0

        if hasattr(value, 'shape'):
            try:
                if value.shape != other_value.shape:
                    return False
            except AttributeError:
                return False

        try:
            ub, lb = value + atol, value - atol
            ub += rtol * np.abs(value)
            lb -= rtol * np.abs(value)
            numeric = True
        
        except:
            numeric = False

        if numeric:
            try:
                res = np.logical_and(ub >= other_value, other_value >= lb)
            except (ValueError, TypeError) as e:
                return False
        else:
            try:
                if (hasattr(value, "shape") and hasattr(other_value, "shape") and value.shape != other_value.shape) \
                        or (hasattr(value, "shape") ^ hasattr(other_value, "shape")):
                    return False
                res = value == other_value
            except (ValueError, TypeError) as e:
                return False

        if isinstance(res, Iterable):
            # handle NaN values in pandas objects
            if isinstance(res, (pd.DataFrame, pd.Series)) and hasattr(value, "isna") and \
                    hasattr(other_value, "isna"):
                res = res | (value.isna() & other_value.isna())
            if isinstance(res, (np.ndarray, pd.DataFrame, pd.Series)):
                res = res.all(axis=None)
            else:
                res = all(res)

        if res:
            return True

        return False


class _AttrValue(Value):
    """
    Wrapper around ``Value`` for checking whether an object has an *attribute* matching a specific
    value.

    Args:
        obj (``object``): the object being checked
        attr (``str``): the attribute of the object being checked
        enforce_type (``bool``, optional): whether to ensure that the satisfying value has the same 
            type as the initial value
        **kwargs: additional keyword arguments passed to the :py:class:`Value<pybryt.Value>` 
            constructor
    """

    _object: Any
    _attr: str

    enforce_type: bool
    """whether to ensure that the satisfying value has the same type as the initial value"""

    def __init__(self, obj: Any, attr: str, enforce_type: bool = False, **kwargs):
        self._object = obj
        self._attr = attr
        self.enforce_type = enforce_type
        val = getattr(obj, attr)
        super().__init__(val, **kwargs)
    
    def __eq__(self, other: Any) -> bool:
        """
        Checks whether this annotation is equal to another object.

        To be equal to a ``_AttrValue``, the other object must also be a ``_AttrValue`` object, have
        equal underlying ``Value`` annotations, have equal source objects, and have the same 
        ``_attr``.

        Args:
            other (``object``): the object to compare to

        Returns:
            ``bool``: whether the objects are equal
        """
        return super().__eq__(other) and self.check_values_equal(self._object, other._object) and \
            self._attr == other._attr and self.enforce_type == other.enforce_type
    
    def check(self, observed_values: List[Tuple[Any, int]]) -> AnnotationResult:
        """
        Checks whether any of the values in ``observed_values`` has an attribute matching the value
        in this annotation.

        Args:
            observed_values (``list[tuple[object, int]]``): a list of tuples of values observed
                during execution and the timestamps of those values
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation based on 
            ``observed_values``
        """
        if self.enforce_type:  # filter out values of wrong type if enforce_type is True
            observed_values = [t for t in observed_values if isinstance(t[0], type(self._object))]
        vals = [t for t in observed_values if hasattr(t[0], self._attr)]
        attrs = [(getattr(obj, self._attr), t) for obj, t in vals]
        res = super().check(attrs)
        try:
            satisfier = vals[attrs.index((res.value, res.timestamp))][0]
        except ValueError:
            satisfier = None
        return AnnotationResult(None, self, value=satisfier, children=[res])


class Attribute(Annotation):
    """
    Annotation for asserting that an object has an attribute value matching that of another object.

    Uses a :py:class:`Value<pybryt.Value>` annotation to check there exists an object in the 
    students' code which as a specific attribute, or a series of attributes, with values matching
    those in the object passed to the constructor. The constructor taks in the object in question
    and the name of the attribute(s) being studied.

    Args:
        obj (``object``): the object being checked
        attrs (``str`` or ``list[str]``): the attribute or attributes that should be checked
        enforce_type (``bool``, optional): whether to ensure that the satisfying value has the same 
            type as the initial value
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.Annotation>` constructor
    """

    _annotations: List[_AttrValue]
    _invariants: List[invariant]
    _atol: Optional[Union[float, int]]
    _rtol: Optional[Union[float, int]]

    enforce_type: bool
    """whether to ensure that the satisfying value has the same type as the initial value"""

    def __init__(self, obj: Any, attrs: Union[str, List[str]], enforce_type: bool = False, **kwargs):
        if isinstance(attrs, str):
            attrs = [attrs]
        if not isinstance(attrs, list) or not all(isinstance(a, str) for a in attrs):
            raise TypeError(f"Invalid type for argument 'attrs': {type(attrs)}")

        name = kwargs.pop("name", None)
        success_message = kwargs.pop("success_message", None)
        failure_message = kwargs.pop("failure_message", None)
        self._annotations = []
        for attr in attrs:
            if not hasattr(obj, attr):
                raise AttributeError(f"{obj} has not attribute '{attr}'")
            self._annotations.append(_AttrValue(obj, attr, enforce_type=enforce_type, **kwargs))
        
        self._invariants = kwargs.pop("invariants", [])
        self._atol = kwargs.pop("atol", None)
        self._rtol = kwargs.pop("rtol", None)

        self.enforce_type = enforce_type
    
        super().__init__(name=name, success_message=success_message, failure_message=failure_message,
            **kwargs)

    @property
    def children(self) -> List[Annotation]:
        return self._annotations
    
    def __eq__(self, other: Any) -> bool:
        """
        Checks whether this annotation is equal to another object.

        To be equal to an ``Attribute``, the other object must also be an ``Attribute`` object and
        have equal child annotations.

        Args:
            other (``object``): the object to compare to

        Returns:
            ``bool``: whether the objects are equal
        """
        return super().__eq__(other) and self.children == other.children

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts this annotation's details to a JSON-friendly dictionary format.

        Output dictionary contains the annotation's name, group, limit number, success message, and
        failure message, as well as an ``invariants`` key with a list of the names of all invariants
        used in this value annotation, a ``atol`` key with this annotation's tolerance, and a key for 
        the attributes being checked. This dictionary does *not* contain the value being tracked.

        Returns:
            ``dict[str, object]``: the dictionary representation of this annotation
        """
        d = super().to_dict()
        d.update({
            "invariants": [inv.__name__ for inv in self._invariants],
            "atol": self._atol,
            "rtol": self._rtol,
            "attributes": [av._attr for av in self._annotations],
            "enforce_type": self.enforce_type,
        })
        return d

    def check(self, observed_values: List[Tuple[Any, int]]) -> AnnotationResult:
        """
        Checks whether any of the values in ``observed_values`` has all of the required attributes,
        each matching the values expected.

        Args:
            observed_values (``list[tuple[object, int]]``): a list of tuples of values observed
                during execution and the timestamps of those values
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation based on 
            ``observed_values``
        """
        results = [v.check(observed_values) for v in self._annotations]        
        return AnnotationResult(None, self, children=results)



"""Tests for PyBryt abstract annotations and annotation results"""

import pytest

from unittest import mock

from pybryt import *

from .utils import *


def test_name_group_limit():
    """
    """
    mfp = generate_memory_footprint()
    val, ts = mfp[2]

    Annotation.reset_tracked_annotations()
    vs = []
    for _ in range(100):
        vs.append(Value(val, name="foo", limit=11))
    
    tracked = Annotation.get_tracked_annotations()
    assert len(tracked) == 11, "Too many tracked annotations"
    assert tracked == vs[:11], "Wrong tracked annotations"
    assert all(v.name == "foo" and v.limit == 11 for v in vs)

    res = vs[-1].check(mfp)
    check_obj_attributes(res, {
        "name": "foo",
        "group": None,
    })

    v1 = Value(mfp[0][0], group="bar")
    v2 = Value(mfp[1][0], group="bar")
    check_obj_attributes(v1, {"group": "bar"})
    check_obj_attributes(v2, {"group": "bar"})

    res = v1.check(mfp)
    print(res.name)
    check_obj_attributes(res, {
        "name": "Annotation 101",
        "group": "bar",
    })


def test_get_reset_tracked_annotations():
    """
    """
    tracked = Annotation.get_tracked_annotations()
    Annotation.reset_tracked_annotations()
    assert len(tracked) == 0

    v1 = Value(1)
    v2 = Value(2)
    assert len(tracked) == 2

    v3 = Value(3)
    assert len(tracked) == 3

    v4 = v3.before(v2)
    assert len(tracked) == 2
    
    assert v1 in tracked
    assert v2 not in tracked
    assert v3 not in tracked
    assert v4 in tracked

    Annotation.reset_tracked_annotations()
    assert len(tracked) == 0


def test_messages():
    """
    """
    mfp = generate_memory_footprint()
    Annotation.reset_tracked_annotations()

    val1, ts1 = mfp[0]
    val2, ts2 = mfp[1]

    v1 = Value(val1, success_message="m1", failure_message="m2")
    v2 = Value(val2)
    
    v = v1.before(v2)
    res = v.check(mfp)

    assert len(res.messages) == 1, "Too many messages"
    assert res.messages[0] == ("m1", 'Annotation 1', True), "Wrong message"

    v.failure_message = "m3"
    res = v.check(mfp)

    assert len(res.messages) == 1, "Too many messages"
    assert res.messages[0] == ("m1", 'Annotation 1', True), "Wrong message"

    res = v.check([])

    assert len(res.messages) == 2, "Wrong number of messages"
    assert res.messages[0] == ("m2", 'Annotation 1', False), "Wrong message"
    assert res.messages[1] == ("m3", 'Annotation 3', False), "Wrong message"

    v2.name = "v2"
    v2.failure_message = "m4"
    res = v.check([])

    assert len(res.messages) == 3, "Wrong number of messages"
    assert res.messages[0] == ("m2", 'Annotation 1', False), "Wrong message"
    assert res.messages[1] == ("m4", "v2", False), "Wrong message"
    assert res.messages[2] == ("m3", 'Annotation 3', False), "Wrong message"


def test_bitwise_ops():
    """
    """
    a1, a2 = Value(1), Value(2)

    assert isinstance(a1 & a2, AndAnnotation), "& operator returns wrong type"
    assert isinstance(a1 | a2, OrAnnotation), "| operator returns wrong type"
    assert isinstance(a1 ^ a2, XorAnnotation), "^ operator returns wrong type"
    assert isinstance(~a1, NotAnnotation), "~ operator returns wrong type"



"""Tests for annotation collections"""

import time
import pytest

from unittest import mock

from pybryt import *
from pybryt.utils import pickle_and_hash

from .utils import *


def test_collection():
    """
    """
    mfp = generate_memory_footprint()
    Annotation.reset_tracked_annotations()

    v1, v2 = Value(mfp[0][0]), Value(mfp[2][0])
    a = Collection(v1, v2, success_message="foo")

    res = a.check(mfp)
    check_obj_attributes(a, {"children__len": 2})
    check_obj_attributes(res, {
        "children__len": 2,
        "satisfied": True,
        "_satisfied": None,
        "annotation": a,
        "timestamp": -1,
        "satisfied_at": 2,
    })

    assert a.to_dict() == {
        "name": "Annotation 3",
        "children": [v1.to_dict(), v2.to_dict()],
        "success_message": "foo",
        "failure_message": None,
        "limit": None,
        "group": None,
        "type": "collection",
        "enforce_order": False,
    }

    a = Collection(v1, v2, enforce_order=True, success_message="foo")
    res = a.check(mfp)
    assert res.satisfied

    v3 = Value(mfp[1][0])
    a.add(v3)
    res = a.check(mfp)
    assert not res.satisfied

    a.remove(v3)
    res = a.check(mfp)
    assert res.satisfied

    a = Collection(v2, v1, enforce_order=True, success_message="foo")
    res = a.check(mfp)
    assert not res.satisfied

    # test errors
    with pytest.raises(ValueError, match="One of the arguments is not an annotation"):
        Collection(v1, 1, v2)
    
    with pytest.raises(TypeError, match="1 is not an annotation"):
        a.add(1)



"""Class for gathering and operating on collections of annotations"""

__all__ = ["Collection"]

from typing import Any, Dict, List, NoReturn, Tuple

from .annotation import Annotation, AnnotationResult


class Collection(Annotation):
    """
    A class for collecting and operating on multiple annotations.

    If ``enforce_order`` is true, this collection will only be satisfied if all of its children are
    satisfied *and* the satisfying timestamps are in non-decreasing order (for those that have them).

    Args:
        *annotations (:py:class:`Annotation<pybryt.annotations.annotation.Annotation>`): the child 
            annotations being operated on
        enforce_order (``bool``, optional): whether to enforce the ordering of annotations as added 
            to this collection
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.annotations.annotation.Annotation>` constructor
    """

    _annotations: List[Annotation]
    """the child annotations that this annotation operates on"""

    enforce_order: bool
    """whether to enforce the ordering of annotations as added to this collection"""

    def __init__(self, *annotations: Annotation, enforce_order: bool = False, **kwargs):
        self._annotations = list(annotations)
        for ann in self._annotations:
            if not isinstance(ann, Annotation):
                raise ValueError("One of the arguments is not an annotation")

        self.enforce_order = enforce_order

        super().__init__(**kwargs)

    @property
    def children(self) -> List[Annotation]:
        """
        ``list[Annotation]``: the child annotations that this annotation operates on
        """
        return self._annotations

    def __eq__(self, other: Any) -> bool:
        """
        Checks whether this annotation is equal to another object.

        For an object to equal a relational annotation, it must also be a relational annotation of
        the same type and have the same child annotations.

        Args:
            other (``object``): the object to compare to

        Returns:
            ``bool``: whether the objects are equal
        """
        return super().__eq__(other) and self.children == other.children and \
            self.enforce_order == other.enforce_order

    def check(self, observed_values: List[Tuple[Any, int]]) -> AnnotationResult:
        """
        Checks that all child annotations are satisfied by the values in ``observed_values``, and
        that the timestamps of the satisfying values occur in non-decreasing order if 
        ``self.enforce_order`` is true.

        Args:
            observed_values (``list[tuple[object, int]]``): a list of tuples of values observed
                during execution and the timestamps of those values
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation based on 
            ``observed_values``
        """
        results = []
        for ann in self.children:
            results.append(ann.check(observed_values))
        
        if self.enforce_order and all(res.satisfied for res in results):
                before = []
                with_timestamp = [res for res in results if res.satisfied_at != -1]
                for i in range(len(with_timestamp) - 1):
                    before.append(with_timestamp[i].satisfied_at < with_timestamp[i + 1].satisfied_at)

                return AnnotationResult(all(before), self, children = results)

        else:
            return AnnotationResult(None, self, children = results)

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts this annotation's details to a JSON-friendly dictionary format.

        Returns:
            ``dict[str, object]``: the dictionary representation of this annotation
        """
        d = super().to_dict()
        d.update({
            "enforce_order": self.enforce_order,
        })
        return d

    def add(self, annotation: Annotation) -> NoReturn:
        """
        Adds an annotation to this collection.

        Args:
            annotation (:py:class:`Annotation<pybryt.annotations.annotation.Annotation>`): the
                annotation to add
        """
        if not isinstance(annotation, Annotation):
            raise TypeError(f"{annotation} is not an annotation")

        self._annotations.append(annotation)
        try:
            self.get_tracked_annotations().remove(annotation)
        except ValueError:  # pragma: no cover
            pass

    def remove(self, annotation: Annotation) -> NoReturn:
        """
        Removes an annotation from this collection.

        Args:
            annotation (:py:class:`Annotation<pybryt.annotations.annotation.Annotation>`): the
                annotation to remove
        """
        if not isinstance(annotation, Annotation):
            raise TypeError(f"{annotation} is not an annotation")

        if annotation not in self._annotations:
            raise ValueError(f"The specified annotation is not part of this collection")

        self._annotations.remove(annotation)



"""Tracing function and controls"""

import re
import linecache
import inspect

from copy import copy
from types import FrameType, FunctionType, ModuleType
from typing import Any, List, Tuple, Callable

from ..utils import make_secret, pickle_and_hash


_COLLECTOR_RET = None
TRACING_VARNAME = "__PYBRYT_TRACING__"
TRACING_FUNC = None


def create_collector(skip_types: List[type] = [type, type(len), ModuleType, FunctionType], addl_filenames: List[str] = []) -> \
        Tuple[Tuple[List[Tuple[Any, int]], List[Tuple[str, str]]], Callable[[FrameType, str, Any], Callable]]:
    """
    Creates a list to collect observed values and a trace function.

    Any types in ``skip_types`` won't be tracked by the trace function. The trace function by 
    default only traces inside IPython but can be set to trace inside specific files using the
    ``addl_filenames`` argument, which should be a list absolute paths to files that should also be
    traced inside of.

    Args:
        skip_types (``list[type]``, optional): object types not to track
        addl_filenames (``list[str]``, optional): filenames to trace inside of in addition to 
            IPython
        
    Returns:
        ``tuple[tuple[list[tuple[object, int]], list[tuple[str, str]]], callable[[frame, str, 
        object], callable]]``: a 2-tuple containing a 2-tuple with the list of tuples of observed 
        objects and their timestamps and the list of call filenames andfunction names, and the trace 
        function
    """
    global _COLLECTOR_RET
    observed = []
    calls = []
    vars_not_found = {}
    hashes = set()
    counter = [0]

    def track_value(val, seen_at=None):
        """
        Tracks a value in ``observed``. Checks that the value has not already been tracked by 
        pickling it and hashing the pickled object and comparing it to ``hashes``. If pickling is
        unsuccessful, the value is not tracked.

        Args:
            val (``object``): the object to be tracked
            seen_at (``int``, optional): an overriding step counter value
        """
        try:
            if type(val) in skip_types:
                return

            h = pickle_and_hash(val)

            if seen_at is None:
                seen_at = counter[0]
            
            if h not in hashes:
                observed.append((copy(val), seen_at))
                hashes.add(h)

        # if something fails, don't track
        except:
            return

    def track_call(frame):
        """
        Tracks a call in ``calls`` as a tuple of ``(filename, function name)``.

        Args:
            frame (``types.FrameType``): the frame of the call
        """
        calls.append((frame.f_code.co_filename, frame.f_code.co_name))

    # TODO: a way to track the cell of execution
    def collect_intermidiate_results(frame: FrameType, event: str, arg: Any):
        """
        Trace function for PyBryt.
        """
        if frame.f_code.co_filename.startswith("<ipython") or frame.f_code.co_filename in addl_filenames:
            counter[0] += 1 # increment student code step counter

        if event == "call":
            track_call(frame)
            return collect_intermidiate_results

        # return if tracking is disabled by a compelxity check
        from .complexity import _TRACKING_DISABLED
        if _TRACKING_DISABLED:
            return collect_intermidiate_results

        name = frame.f_code.co_filename + frame.f_code.co_name

        if frame.f_code.co_filename.startswith("<ipython") or frame.f_code.co_filename in addl_filenames:
            if event == "line" or event == "return":

                line = linecache.getline(frame.f_code.co_filename, frame.f_lineno)
                tokens = set("".join(char if char.isalnum() or char == '_' else "\n" for char in line).split("\n"))
                for t in "".join(char if char.isalnum() or char == '_' or char == '.' else "\n" for char in line).split("\n"):
                    tokens.add(t)
                tokens = sorted(tokens) # sort for stable ordering
                
                for t in tokens:
                    if "." in t:
                        try:
                            val = eval(t, frame.f_globals, frame.f_locals)
                            track_value(val)
                        except:
                            pass

                    else:
                        if t in frame.f_locals:
                            val = frame.f_locals[t]
                            track_value(val)
                                
                        elif t in frame.f_globals:
                            val = frame.f_globals[t]
                            track_value(val)
                
                # for tracking the results of an assignment statement
                m = re.match(r"^\s*(\w+)(\[[^\]]\]|(\.\w+)+)*\s=.*", line)
                if m:
                    if name not in vars_not_found:
                        vars_not_found[name] = []
                    vars_not_found[name].append((m.group(1), counter[0]))

            if event == "return":
                track_value(arg)

        elif (frame.f_back.f_code.co_filename.startswith("<ipython") or \
                frame.f_back.f_code.co_filename in addl_filenames) and event == "return":
            track_value(arg)

        if event == "return" and name in vars_not_found:
            varnames = vars_not_found.pop(name)
            for t, step in varnames:
                if t in frame.f_locals:
                    val = frame.f_locals[t]
                    track_value(val, step)

                elif t in frame.f_globals:
                    val = frame.f_globals[t]
                    track_value(val, step)

        return collect_intermidiate_results

    _COLLECTOR_RET = (observed, counter, collect_intermidiate_results)
    return (observed, calls), collect_intermidiate_results


def _get_tracing_frame():
    """
    Returns the frame that is being traced by looking for the ``__PYBRYT_TRACING__`` global variable.

    Returns:
        the frame being traced or ``None`` of no tracing is occurring
    """
    frame = inspect.currentframe()
    while frame is not None:
        if TRACING_VARNAME in frame.f_globals and frame.f_globals[TRACING_VARNAME]:
            return frame
        frame = frame.f_back
    return None


def tracing_off(frame=None, save_func=True):
    """
    Turns off PyBryt's tracing if tracing is occurring in this call stack. If PyBryt is not tracing,
    takes no action.

    This method can be used in students' notebooks to include code that shouldn't be traced as part
    of the submission, e.g. demo code or ungraded code. In the example below, the call that creates
    ``x2`` is traced but the one to create ``x3`` is not.

    .. code-block:: python

        def pow(x, a):
            return x ** a

        x2 = pow(x, 2)

        pybryt.tracing_off()
        x3 = pow(x, 3)
    """
    global TRACING_FUNC
    frame = _get_tracing_frame() if frame is None else frame
    if frame is None:
        return
    if save_func:
        TRACING_FUNC = frame.f_trace
    vn = f"sys_{make_secret()}"
    exec(f"import sys as {vn}\n{vn}.settrace(None)", frame.f_globals, frame.f_locals)


def tracing_on(frame=None, tracing_func=None):
    """
    Turns tracing on if PyBryt was tracing the call stack. If PyBryt is not tracing or
    :py:meth:`tracing_off<pybryt.tracing_off>` has not been called, no action is taken.

    This method can be used in students' notebooks to turn tracing back on after deactivating tracing
    for ungraded code In the example below, ``x4`` is traced because ``tracing_on`` is used after
    ``tracing_off`` and the creation of ``x3``.

    .. code-block:: python

        def pow(x, a):
            return x ** a

        x2 = pow(x, 2)

        pybryt.tracing_off()
        x3 = pow(x, 3)
        pybryt.tracing_on()

        x4 = pow(x, 4)
    """
    global TRACING_FUNC
    frame = _get_tracing_frame() if frame is None else frame
    if frame is None or (TRACING_FUNC is None and tracing_func is None):
        return
    if TRACING_FUNC is not None and tracing_func is None:
        tracing_func = TRACING_FUNC
    vn = f"cir_{make_secret()}"
    vn2 = f"sys_{make_secret()}"
    frame.f_globals[vn] = tracing_func
    exec(f"import sys as {vn2}\n{vn2}.settrace({vn})", frame.f_globals, frame.f_locals)
    frame.f_trace = tracing_func


class no_tracing:
    """
    A context manager for turning tracing off for a block of code in a submission.

    If PyBryt is tracing code, any code inside this context will not be traced for values in memory.
    If PyBryt is not tracing, no action is taken.

    .. code-block:: python

        with pybryt.no_tracing():
            # this code is not traced
            foo(1)
        
        # this code is traced
        foo(2)
    """

    def __enter__(self):
        tracing_off()

    def __exit__(self, exc_type, exc_value, traceback):
        tracing_on()
        return False



"""Annotation classes for complexity assertions"""

__all__ = ["ComplexityAnnotation", "TimeComplexity"]

from typing import Any, List, Tuple

from . import complexities as cplx
from ..annotation import Annotation, AnnotationResult
from ...execution import TimeComplexityResult


EPS = 1e-6 # a value to set a slight preference for simpler methods


class ComplexityAnnotation(Annotation):
    """
    Abstract base class for annotations that assert a condition on the complexity of a student's code.

    All complexity annotations should inherit from this class, which defines a constructor that
    takes in a complexity class and any necessary keyword arguments. Note that all compelxity
    annotations require a ``name`` keyword argument so that they can be matched with their results
    from the student's memory footprint.

    Args:
        complexity (:py:class:`complexity<pybryt.complexities.complexity>`): the complexity class
            being asserted
        addl_complexities(``list[complexity]``): additional custom complexity classes to consider
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.Annotation>` constructor
    """

    def __init__(self, complexity: cplx.complexity, addl_complexities=[], **kwargs):
        if "name" not in kwargs:
            raise ValueError("Complexity annotations require a 'name' kwarg")
        if complexity not in cplx.complexity_classes and not isinstance(complexity, cplx.complexity):
            raise ValueError(f"Invalid valid for argument 'complexity': {complexity}")
        self.complexity = complexity
        self.addl_complexities = addl_complexities
        super().__init__(**kwargs)

    @property
    def children(self) -> List[Annotation]:
        """
        ``list[Annotation]``: the child annotations of this annotation. If this annotation has no 
        children, an empty list.
        """
        return []
    
    def __eq__(self, other: Any) -> bool:
        """
        Checks whether this annotation is equal to another object.

        A complexity annotation equals another object if it is also a complexity annotation of the
        same type, has the same name, and has the same complexity assertion.

        Args:
            other (``object``): the object to compare to

        Returns:
            ``bool``: whether the objects are equal
        """
        return super().__eq__(other) and self.complexity is other.complexity


class TimeComplexity(ComplexityAnnotation):
    """
    Annotation for asserting the time complexity of a block of student code.

    Time complexity here is defined as the number of execution steps taken while executing the
    code block, which is determined using the number of calls to PyBryt's trace function. Use the
    :py:obj:`check_time_complexity<pybryt.check_time_complexity>` context manager to check time
    complexity in student's code. The ``name`` of this annotation should be the same as the ``name``
    passed to the context manager or this annotation will not be able to find the results of the 
    check.
    """

    def check(self, observed_values: List[Tuple[Any, int]]) -> AnnotationResult:
        """
        Checks the time complexity of a block of student code and returns a results object.

        Finds all instances of the :py:class:`TimeComplexityResult<pybryt.execution.TimeComplexityResult>` 
        class in the student's memory footprint and selects all those with matching names. Collects
        the timing data from each into a dictionary and runs each complexity class in
        :py:obj:`complexities.complexity_classes<pybryt.complexities.complexity_classes>` against
        this data. Returns a result indicating whether the closest-matched complexity class was the
        one asserted in this annotation's ``complexity`` field. The ``value`` of the result object
        is set to the matching complexity class.

        Args:
            observed_values (``list[tuple[object, int]]``): a list of tuples of values observed
                during execution and the timestamps of those values
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation based on 
            ``observed_values``
        """
        if self.complexity not in cplx.complexity_classes:
            self.addl_complexities.insert(0, self.complexity)

        complexity_data = {}
        for v, ts in observed_values:
            if not isinstance(v, TimeComplexityResult) or v.name != self.name:
                continue

            complexity_data[v.n] = v.stop - v.start

        best_cls, best_res = None, None
        complexities = cplx.complexity_classes + self.addl_complexities
        for cplxy in complexities:
            res = cplxy(complexity_data)

            if best_res is None or res < best_res - EPS:
                best_cls, best_res = cplxy, res

        return AnnotationResult(best_cls is self.complexity, self, value=best_cls)



"""Submission execution internals for PyBryt"""

__all__ = ["check_time_complexity", "no_tracing"]

import os
import dill
import nbformat

from nbconvert.preprocessors import ExecutePreprocessor
from copy import deepcopy
from tempfile import mkstemp
from typing import Any, List, Tuple, Optional
from textwrap import dedent

from .complexity import check_time_complexity, TimeComplexityResult
from .tracing import (
    create_collector, _get_tracing_frame, no_tracing, tracing_off, tracing_on, TRACING_VARNAME
)
from ..preprocessors import IntermediateVariablePreprocessor
from ..utils import make_secret


NBFORMAT_VERSION = 4


def execute_notebook(nb: nbformat.NotebookNode, nb_path: str, addl_filenames: List[str] = [], 
        output: Optional[str] = None) -> Tuple[int, List[Tuple[Any, int]], List[Tuple[str, str]], \
        nbformat.NotebookNode]:
    """
    Executes a submission using ``nbconvert`` and returns the memory footprint.

    Takes in a notebook object and preprocesses it before running it through the 
    ``nbconvert.ExecutePreprocessor`` to execute it. The notebook writes the memory footprint, a 
    list of observed values and their timestamps, to a file, which is loaded using ``dill`` by this
    function. Errors during execution are ignored, and the executed notebook can be written to a 
    file using the ``output`` argument.

    Args:
        nb (``nbformat.NotebookNode``): the notebook to be executed
        nb_path (``str``): path to the notebook ``nb``
        addl_filenames (``list[str]``, optional): a list of additional files to trace inside
        output (``str``, optional): a file path at which to write the executed notebook

    Returns:
        ``tuple[int, list[tuple[object, int]], list[tuple[str, str]], nbformat.NotebookNode]``: the 
        number of execution steps, the memory footprint, the list of function calls, and the 
        executed notebook
    """
    nb = deepcopy(nb)
    preprocessor = IntermediateVariablePreprocessor()
    nb = preprocessor.preprocess(nb)

    secret = make_secret()
    _, observed_fp = mkstemp()
    nb_dir = os.path.abspath(os.path.split(nb_path)[0])

    first_cell = nbformat.v4.new_code_cell(dedent(f"""\
        import sys
        from pybryt.execution import create_collector
        cir_results_{secret}, cir = create_collector(addl_filenames={addl_filenames})
        sys.settrace(cir)
        {TRACING_VARNAME} = True
        %cd {nb_dir}
    """))

    last_cell = nbformat.v4.new_code_cell(dedent(f"""\
        sys.settrace(None)
        import dill
        from pybryt.utils import filter_picklable_list
        filter_picklable_list(cir_results_{secret}[0])
        with open("{observed_fp}", "wb+") as f:
            dill.dump(cir_results_{secret}, f)
    """))

    nb['cells'].insert(0, first_cell)
    nb['cells'].append(last_cell)

    ep = ExecutePreprocessor(timeout=1200, allow_errors=True)

    ep.preprocess(nb)

    if output:
        with open(output, "w+") as f:
            nbformat.write(nb, f)

    with open(observed_fp, "rb") as f:
        observed, calls = dill.load(f)

    os.remove(observed_fp)

    n_steps = max([t[1] for t in observed])

    return n_steps, observed, calls, nb



""""""

import sys
import inspect
import numpy as np

from unittest import mock

from pybryt import *
from pybryt.execution import create_collector, tracing_off, tracing_on, TRACING_VARNAME

from .utils import generate_mocked_frame


__PYBRYT_TRACING__ = False


def test_trace_function():
    """
    """
    np.random.seed(42)
    
    tracked_filepath = "/path/to/tracked/file.py"

    frame = generate_mocked_frame("<ipython-abc123>", "foo", 3)
    (observed, calls), cir = create_collector(addl_filenames=[tracked_filepath])

    arr = np.random.uniform(-100, 100, size=(100, 100))
    cir(frame, "return", arr)
    assert len(observed) == 1
    assert np.allclose(observed[0][0], arr)
    assert observed[0][1] == 1

    # test value in skip_types
    cir(frame, "return", type(1))
    assert len(observed) == 1

    # test pickling error
    with mock.patch("dill.dumps") as mocked_dumps:
        mocked_dumps.side_effect = Exception()
        cir(frame, "return", 1)
    
    assert len(observed) == 1

    # test call event
    assert len(calls) == 0
    cir(frame, "call", None)
    assert len(calls) == 1
    assert calls[0] == ("<ipython-abc123>", "foo")

    frame = generate_mocked_frame(
        "<ipython-abc123>", "foo", 3, {"data": arr}
    )

    # check line processing working
    with mock.patch("linecache.getline") as mocked_linecache:

        # check eval call for attributes
        mocked_linecache.return_value = "data.T"
        cir(frame, "line", None)
        assert len(observed) == 2
        assert np.allclose(observed[1][0], arr.T)
        assert observed[1][1] == 5

        # check failed eval call for attributes
        mocked_linecache.return_value = "data.doesnt_exist"
        cir(frame, "line", None)
        assert len(observed) == 2

        # check looking in frame locals + globals
        frame.f_globals["more_data"] = np.random.uniform(-100, 100, size=(100, 100))
        frame.f_locals["more_data"] = np.random.uniform(-100, 100, size=(100, 100))
        mocked_linecache.return_value = "more_data"
        cir(frame, "line", None)
        assert len(observed) == 3
        assert np.allclose(observed[2][0], frame.f_locals["more_data"])
        assert observed[2][1] == 7

        # check that we track assignment statements on function return
        mocked_linecache.return_value = "even_more_data = more_data ** 2"
        cir(frame, "line", None)
        assert len(observed) == 3

        mocked_linecache.return_value = "even_more_data_2 = more_data ** 3"
        cir(frame, "line", None)
        assert len(observed) == 3

        frame.f_locals["even_more_data"] = frame.f_locals["more_data"] ** 2
        frame.f_globals["even_more_data_2"] = frame.f_locals["more_data"] ** 3
        mocked_linecache.return_value = ""
        cir(frame, "return", None)
        assert len(observed) == 6
        assert observed[3][0] is None
        assert observed[3][1] == 10
        assert np.allclose(observed[4][0], frame.f_locals["more_data"] ** 2)
        assert observed[4][1] == 8
        assert np.allclose(observed[5][0], frame.f_locals["more_data"] ** 3)
        assert observed[5][1] == 9

        # check that skip_types respected
        frame.f_locals["none_type"] = type(None)
        mocked_linecache.return_vaue = "none_type"
        cir(frame, "line", None)
        assert len(observed) == 6

        # check that addl_filenames respected
        frame = generate_mocked_frame(tracked_filepath, "bar", 100, f_back=frame)
        frame.f_locals["data"] = arr
        mocked_linecache.return_value = "arr = -1 * data"
        cir(frame, "line", None)
        frame.f_locals["arr"] = -1 * arr
        cir(frame, "return", None) # run a return since arr shows up in vars_not_found
        assert len(observed) == 7
        assert np.allclose(observed[6][0], -1 * arr)
        assert observed[6][1] == 13
    
    # check that IPython child frame return values are tracked
    frame = generate_mocked_frame("/path/to/file.py", "bar", 100, f_back=frame)
    cir(frame, "line", None)
    assert len(observed) == 7

    cir(frame, "return", np.exp(arr))
    assert len(observed) == 8
    assert np.allclose(observed[7][0], np.exp(arr))
    assert observed[7][1] == 13

    assert len(calls) == 1
    frame = generate_mocked_frame("/path/to/foo.py", "bar", 100)
    cir(frame, "call", None)
    assert len(calls) == 2
    assert calls[0] == ("<ipython-abc123>", "foo")
    assert calls[1] == ("/path/to/foo.py", "bar")


def test_tracing_control():
    """
    """
    global __PYBRYT_TRACING__
    trace = lambda frame, event, arg: trace

    __PYBRYT_TRACING__ = True
    with mock.patch("sys.settrace") as mocked_settrace:
        tracing_off()
        mocked_settrace.assert_called()

    with mock.patch("sys.settrace") as mocked_settrace:
        tracing_on()
        if sys.gettrace() is not None:
            mocked_settrace.assert_called()

    __PYBRYT_TRACING__ = False
    with mock.patch("sys.settrace") as mocked_settrace:
        tracing_off()
        mocked_settrace.assert_not_called()

    with mock.patch("sys.settrace") as mocked_settrace:
        tracing_on()
        mocked_settrace.assert_not_called()


def test_tracing_context_manager():
    """
    """
    with mock.patch("pybryt.execution.tracing.tracing_off") as mocked_off, \
            mock.patch("pybryt.execution.tracing.tracing_on") as mocked_on:
        with no_tracing():
            mocked_off.assert_called()
            mocked_on.assert_not_called()
        mocked_on.assert_called()



"""Invariants for value annotations"""

import numpy as np

from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Any, List, Optional, Union


class invariant(ABC):
    """
    Abstract base class for invariants. 
    
    All subclasses should implement the :py:meth:`run<invariant.run>` static method for generating 
    values that this invariant accepts as "correct". Invariants have a custom ``__new__`` method 
    that returns the value of calling the :py:meth:`run<invariant.run>` method, making them 
    function as callables.
    """

    @staticmethod
    def __new__(cls, *args, **kwargs):
        return cls.run(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def run(values: List[Any], **kwargs) -> List[Any]:
        """
        Returns a list of values that this invariant accepts as correct.

        Takes in a list of acceptable values from a :py:class:`Value<pybryt.Value>` annotation and
        returns a list of values that would evaluate as "the same" under the conditions of this 
        invariant.

        For example, if ``values`` as a list with a single element, a numpy matrix, and the 
        invariant was matrix transposition, this method would return a length-2 list containing the
        original matrix and its transpose.

        Args:
            values (``list[object]``): acceptable values, either from the initial constructor call
                of the annotation or from the results of other invariants
            kwargs: additional keyword arguments
        
        Returns:
            ``list[object]``: the values that would evaluate as "the same" under the conditions 
            of this invariant
        """
        ... # pragma: no cover


class string_capitalization(invariant):
    """
    An invariant that compares strings ignoring the capitalization of letters.
    """

    @staticmethod
    def run(values: List[Any]) -> List[Any]:
        """
        Returns a list of values in which all strings have been lowercased.

        Args:
            values (``list[object]``): acceptable values, either from the initial constructor call
                of the annotation or from the results of other invariants
        
        Returns:
            ``list[object]``: the transformed values
        """
        ret = []
        for v in values:
            if not isinstance(v, str):
                ret.append(v)
            else:
                ret.append(v.lower())
        return ret


class matrix_transpose(invariant):
    """
    An invariant that compares 2-dimensional arrays ignoring transposition.
    """

    @staticmethod
    def run(values: List[Any]) -> List[Any]:
        """
        Returns a list of values in which all 2D iterables have been converted to NumPy arrays and
        have had their transpose added

        Args:
            values (``list[object]``): acceptable values, either from the initial constructor call
                of the annotation or from the results of other invariants
        
        Returns:
            ``list[object]``: the transformed values
        """
        ret = []
        for v in values:
            if isinstance(v, np.ndarray):
                ret.append(v)
                ret.append(v.T)
            elif isinstance(v, Iterable) and not isinstance(v, str):
                try:
                    arr = np.array(v)
                    ret.append(arr)
                    ret.append(arr.T)
                except:
                    ret.append(v)
            else:
                ret.append(v)
        return ret


class list_permutation(invariant):
    """
    An invariant that compares iterables (except strings) ignoring ordering, using ``sorted``.
    """

    @staticmethod
    def run(values: List[Any]) -> List[Any]:
        """
        Returns a list of values in which all iterables have been sorted.

        Args:
            values (``list[object]``): acceptable values, either from the initial constructor call
                of the annotation or from the results of other invariants
        
        Returns:
            ``list[object]``: the transformed values
        """
        ret = []
        for v in values:
            if isinstance(v, np.ndarray):
                ret.append(np.sort(v))
            elif isinstance(v, Iterable) and not isinstance(v, str):
                ret.append(sorted(v))
            else:
                ret.append(v)
        return ret



"""Tests for PyBryt execution internals"""

import re
import dill
import pathlib
import random
import tempfile
import nbformat
import numpy as np

from unittest import mock

from pybryt.execution import execute_notebook


def generate_test_notebook():
    """
    """
    nb = nbformat.v4.new_notebook()
    nb.cells.append(nbformat.v4.new_code_cell(
        "import numpy as np\nimport pandas as pd\nimport matplotlib.pyplot as plt\n%matplotlib inline"
        ))
    nb.cells.append(nbformat.v4.new_code_cell(
        "np.random.seed(42)\nx = np.random.uniform(size=1000)\ny = np.random.normal(size=1000)"
    ))
    nb.cells.append(nbformat.v4.new_code_cell("df = pd.DataFrame({'x': x, 'y': y})"))
    return nb


def test_notebook_execution():
    """
    """
    random.seed(42)
    nb = generate_test_notebook()

    observed_fn = str(pathlib.Path(__file__).parent.parent / 'files' / 'expected_observed.pkl')
    with open(observed_fn, "rb") as f:
        expected_observed = dill.load(f)

    with tempfile.NamedTemporaryFile("w+") as ntf:
        with tempfile.NamedTemporaryFile(delete=False) as observed_ntf:
            with mock.patch("pybryt.execution.mkstemp") as mocked_tempfile:
                mocked_tempfile.return_value = (None, observed_ntf.name)

                n_steps, observed, calls, _ = execute_notebook(nb, "", output=ntf.name)
                assert len(ntf.read()) > 0
                assert n_steps == max(t[1] for t in observed)
                assert isinstance(calls, list) and isinstance(calls[0], tuple) and \
                    isinstance(calls[0][0], str) and isinstance(calls[0][1], str)



"""Tests for PyBryt value annotations"""

import time
import pytest

from unittest import mock

from pybryt import *
from pybryt.utils import pickle_and_hash

from .utils import *


def test_value_annotation():
    """
    """
    mfp = generate_memory_footprint()
    Annotation.reset_tracked_annotations()

    seen = {}
    for val, ts in mfp:
        v = Value(val)
        res = v.check(mfp)

        assert repr(res) == "AnnotationResult(satisfied=True, annotation=pybryt.Value)"

        h = pickle_and_hash(val)

        # check attributes of BeforeAnnotation and AnnotationResult
        check_obj_attributes(v, {"children__len": 0})
        check_obj_attributes(res, {
            "children": [],
            "satisfied": True,
            "_satisfied": True,
            "annotation": v,
            "timestamp": seen[h] if h in seen else ts,
            "value": val,
        })

        if h not in seen:
            seen[h] = ts

    v = Value(-1) # does not occur in mfp
    res = v.check(mfp)

    assert v.to_dict() == {
        "name": "Annotation 11",
        "children": [],
        "success_message": None,
        "failure_message": None,
        "limit": None,
        "group": None,
        "invariants": [],
        "atol": None,
        "rtol": None,
        "type": "value",
    }
    assert repr(res) == "AnnotationResult(satisfied=False, annotation=pybryt.Value)"

    # check __repr__
    assert repr(v) == "pybryt.Value", "wrong __repr__"

    # check attributes of BeforeAnnotation and AnnotationResult
    check_obj_attributes(v, {"children__len": 0})
    check_obj_attributes(res, {
        "children": [],
        "satisfied": False,
        "_satisfied": False,
        "annotation": v,
        "timestamp": -1,
        "value": None,
    })

    # test pickling error
    with mock.patch("dill.dumps") as mocked_dumps:
        mocked_dumps.side_effect = Exception()
        with pytest.raises(ValueError):
            v = Value(-1)

    # test with invariants
    s = mfp[-1][0]
    v = Value(s.upper(), invariants=[invariants.string_capitalization])
    res = v.check(mfp)
    assert res.satisfied


def test_attribute_annotation():
    """
    """
    mfp = generate_memory_footprint()
    Annotation.reset_tracked_annotations()
    val, ts = mfp[0]

    v = Attribute(val, "T")
    res = v.check(mfp)

    # check attributes of BeforeAnnotation and AnnotationResult
    check_obj_attributes(v, {"children__len": 1})
    check_obj_attributes(res, {
        "children__len": 1,
        "satisfied": True,
        "_satisfied": None,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": ts,
        "value": val,
    })

    assert v.to_dict() == {
        "name": "Annotation 2",
        "children": [
            {
                'name': 
                'Annotation 1', 
                'group': None, 
                'limit': None, 
                'success_message': None, 
                'failure_message': None, 
                'children': [], 
                'invariants': [], 
                'atol': None,
                'rtol': None,
                "type": None,
            }
        ],
        "success_message": None,
        "failure_message": None,
        "limit": None,
        "group": None,
        "invariants": [],
        "atol": None,
        "rtol": None,
        "type": "attribute",
        "attributes": ['T'],
        "enforce_type": False,
    }

    # check enforce type
    class Foo:
        T = val.T

    mfp2 = [(Foo(), 1)]
    res = v.check(mfp2)
    assert res.satisfied

    v = Attribute(val, "T", enforce_type=True)
    res = v.check(mfp2)
    assert not res.satisfied

    res = v.check(mfp + mfp2)
    assert res.satisfied

    # check error raising
    with pytest.raises(TypeError):
        Attribute(val, ["T", 1])
    
    with pytest.raises(TypeError):
        Attribute(val, 1)

    with pytest.raises(AttributeError):
        Attribute(val, "foo")


"""Intermediate variable preprocessor for PyBryt submissions"""

import ast
import astunparse
import nbformat

from IPython.core.inputtransformer2 import TransformerManager

from ..utils import make_secret, notebook_to_string


class UnassignedVarWrapper(ast.NodeTransformer):
    """
    AST node transformer that creates intermediate variables for any calls in nested expressions

    Transforms an AST such that every ``Call`` or ``BinOp`` which does not have an ``Assign`` as its 
    parent is added to the AST in the closest parent node with a ``body`` as an intermediate v
    ariable. The node is replaced with a ``Name`` of the new variable name. Any nodes who have a 
    parent node that is an instance of any node type in ``UnassignedVarWrapper._skip_node_types`` 
    is not transformed.

    Attrs:
        insertions (``list[tuple[ast.Node, str, int, ast.Node]]``): a tuple of insertions to make in 
            a node's body; for the tuple ``(parent, attr, idx, node)``, performing the insertion is 
            done by running ``getattr(parent, attr).insert(idx, node)``
    """

    _skip_node_types = [ast.ListComp, ast.SetComp, ast.GeneratorExp, ast.DictComp, ast.Lambda, ast.withitem]

    def __init__(self):
        super().__init__()
        self.insertions = [] # list of (parent, attr, idx, node, orig_node, name_node)
        self.overwrites = [] # list of (parent, attr, index, node)

    @staticmethod
    def add_parents(root):
        """
        Adds a ``parent`` field to each node in the AST rooted at ``root`` pointed to the direct
        parent of that node.

        Args:
            root (``ast.Node``): the root of the AST
        """
        for node in ast.walk(root):
            for child in ast.iter_child_nodes(node):
                child.parent = node

    @staticmethod
    def get_varname():
        """
        Returns a random, valid Python identifier as a string

        Returns:
            ``str``: the variable name
        """
        return f"var_{make_secret()}"

    def visit(self, root, *args, top_level=False, **kwargs):
        """
        Visits each node in the tree to transform all ``Call`` nodes and then performs the node
        insertions specified in ``self.insertions``. Arguments and return values are the same as 
        ``ast.NodeTransformer``.
        """
        self.root = root
        ret = super().visit(root, *args, **kwargs)
        if top_level:
            for parent, attr, idx, node in self.overwrites:
                if idx is None:
                    setattr(parent, attr, node)
                else:
                    getattr(parent, attr)[idx] = node
            self.fix_bool_ops()
            for parent, attr, idx, node, _, body_child, _ in self.insertions:
                getattr(parent, attr).insert(getattr(parent, attr).index(body_child), node)
        return ret

    def transform_unassigned_node(self, node):
        """
        Transforms an unassigned node into an intermediate variable to be inserted in the closest
        ancestor node with a body. Updates ``self.insertions`` with the insertion needed to be
        performed after ``super().visit`` finishes.

        Args:
            node (``ast.Node``): the node to be transformed; should have its ``parent`` attribute
                set by ``self.add_parents``

        Returns:
            ``ast.Node``: an untransformed node if no transformation was required or the new 
            ``Name`` node to be inserted into the AST if transformation was required.
        """
        for n in ast.iter_child_nodes(node):
            self.visit(n)

        if not isinstance(node.parent, ast.Assign) and not isinstance(node.parent, ast.Module):
            vn = self.get_varname()
            curr = node.parent

            for attr in dir(curr):
                if getattr(curr, attr) == node:
                    overwrite_attr = attr
                    overwrite_index = None
                    break
                elif isinstance(getattr(curr, attr), list) and node in getattr(curr, attr):
                    overwrite_attr = attr
                    overwrite_index = getattr(curr, attr).index(node)

            body_child = None
            while not isinstance(curr, ast.Module):
                if hasattr(curr.parent, "body") and not (hasattr(curr.parent, "test") and \
                        curr.parent.test == curr) and body_child is None and \
                        not (isinstance(curr.parent, ast.For) and curr.parent.iter == curr):
                    body_child = curr

                # don't perform if in a comprehension
                if any(isinstance(curr, t) for t in type(self)._skip_node_types):
                    return node

                curr = curr.parent

            curr = body_child
            is_else = False
            try:
                idx = curr.parent.body.index(curr)
            except ValueError:
                if isinstance(curr.parent, ast.If):
                    idx = curr.parent.orelse.index(curr)
                    is_else = True
                else: # pragma: no cover
                    raise

            curr = curr.parent
            new_assign = ast.Assign([ast.Name(vn, ast.Store())], node)
            new_assign.parent = curr

            new_name = ast.Name(vn, ast.Load())
            new_name.parent = node.parent
            node.parent = new_assign

            if is_else:
                self.insertions.append((curr, "orelse", idx, new_assign, node, body_child, new_name))
            else:
                self.insertions.append((curr, "body", idx, new_assign, node, body_child, new_name))

            self.visit(new_assign)

            self.overwrites.append((new_name.parent, overwrite_attr, overwrite_index, new_name))

            return new_name

        return node

    def fix_bool_ops(self):
        """
        """
        for i, (parent, attr, idx, node, orig_node, body_child, name_node) in enumerate(self.insertions):
            if isinstance(name_node.parent, ast.BoolOp):
                try:
                    idx = name_node.parent.values.index(orig_node)
                except ValueError:
                    idx = name_node.parent.values.index(name_node)
                if idx > 0:
                    conds = name_node.parent.values[:idx]                    
                    op = name_node.parent.op
                    new_bool = ast.BoolOp(op, conds)
                    if isinstance(op, ast.Or):
                        new_bool = ast.UnaryOp(ast.Not(), new_bool)
                    new_if = ast.If(new_bool, [node], [])
                    new_bool.parent = new_if
                    node.parent = new_if
                    new_if.parent = parent
                    self.insertions[i] = (parent, attr, idx, new_if, orig_node, body_child, name_node)

    def visit_Call(self, node):
        """
        Transforms all ``Call`` nodes if necessary.
        """
        return self.transform_unassigned_node(node)

    def visit_BinOp(self, node):
        """
        Transforms all ``BinOp`` nodes if necessary.
        """
        return self.transform_unassigned_node(node)


class IntermediateVariablePreprocessor():
    """
    Preprocessor for inserting intermediate variables in a notebook to assist in tracing.
    """

    def preprocess(self, nb):
        """
        Preprocesses a notebook by inserting intermediate variables.

        Iterates through the cells in a notebook, converting them to ASTs and using
        :py:class:`UnassignedVarWrapper<pybryt.preprocessors.intedmediate_variablea.UnassignedVarWrapper>`
        to processor the code and ``astunparse.unparse`` to turn the result back into a string. 
        Updates the notebook inplace.

        Args:
            nb (``nbformat.NotebookNode``): the notebook to be preprocessed

        Returns:
            ``nbformat.NotebookNode``: the updated notebook
        """
        transformer_mgr = TransformerManager()
        for cell in nb['cells']:
            if cell['cell_type'] == 'code':
                code = cell['source']
                code = transformer_mgr.transform_cell(code)
                tree = ast.parse(code)
                transformer = UnassignedVarWrapper()
                transformer.add_parents(tree)
                tree = transformer.visit(tree, top_level=True)
                tree = ast.fix_missing_locations(tree)
                code = astunparse.unparse(tree)
                cell['source'] = code

        return nb



"""Complexity classes for complexity annotations"""

import numpy as np

from abc import ABC, abstractmethod
from typing import Dict


class complexity(ABC):
    """
    Abstract base class for a complexity. Subclassses should implement the ``transform_n`` method,
    which transforms the input lengths array so that least squares can be used. If needed, the
    ``transform_t`` method can also be overwritten to transform the step counter values.

    The architecture for these and the algorithm for determining the optimal complexity is borrowed
    from https://github.com/pberkes/big_O.
    """

    @staticmethod
    def __new__(cls, complexity_data: Dict[int, int]) -> float:
        return cls.run(complexity_data)

    @staticmethod
    @abstractmethod
    def transform_n(n: np.ndarray) -> np.ndarray:
        """
        Transforms the array of input lengths for performing least squares.

        Args:
            n (``np.ndarray``): the array of input lengths
        
        Returns:
            ``np.ndarray``: the transformed array of input lengths
        """
        ... # pragma: no cover

    @staticmethod
    def transform_t(t: np.ndarray) -> np.ndarray:
        """
        Transforms the array of timings for performing least squares.

        Args:
            n (``np.ndarray``): the array of timings
        
        Returns:
            ``np.ndarray``: the transformed array of timings
        """
        return t

    @classmethod
    def run(cls, complexity_data: Dict[int, int]) -> float:
        """
        Returns the sum of residuals by performing least squares on the input length data and timings.

        Uses ``transform_n`` and ``transform_t`` to transform the data from ``complexity_data``, which
        is a dictionary mapping input lengths to step counts. Performs least squares on the resulting
        arrays and returns the sum of residuals.

        Args:
            complexity_data (``dict[int, int]``): the complexity information

        Returns:
            ``float``: the sum of residuals from least squares
        """
        ns, ts = [], []
        for n, t in complexity_data.items():
            ns.append(n)
            ts.append(t)
        
        n = np.array(ns, dtype=int)
        t = np.array(ts, dtype=int)

        try:
            n = cls.transform_n(n)
            t = cls.transform_t(t)

            _, resid, _, _ = np.linalg.lstsq(n, t, rcond=-1)
            if len(resid) == 0:
                return np.inf
            return resid[0]
        except:
            return np.inf


class constant(complexity):
    """
    Complexity class for constant time: :math:`\\mathcal{O}(1)`
    """

    @staticmethod
    def transform_n(n: np.ndarray) -> np.ndarray:
        return np.ones((len(n), 1))


class logarithmic(complexity):
    """
    Complexity class for logarithmic time: :math:`\\mathcal{O}(\\log n)`
    """

    @staticmethod
    def transform_n(n: np.ndarray) -> np.ndarray:
        return np.vstack((np.ones(len(n)), np.log2(n))).T


class linear(complexity):
    """
    Complexity class for linear time: :math:`\\mathcal{O}(n)`
    """

    @staticmethod
    def transform_n(n: np.ndarray) -> np.ndarray:
        return np.vstack((np.ones(len(n)), n)).T


class linearithmic(complexity):
    """
    Complexity class for linearithmic time: :math:`\\mathcal{O}(n \\log n)`
    """

    @staticmethod
    def transform_n(n: np.ndarray) -> np.ndarray:
        return np.vstack((np.ones(len(n)), n * np.log2(n))).T


class quadratic(complexity):
    """
    Complexity class for quadratic time: :math:`\\mathcal{O}(n^2)`
    """

    @staticmethod
    def transform_n(n: np.ndarray) -> np.ndarray:
        return np.vstack((np.ones(len(n)), n * n)).T


class cubic(complexity):
    """
    Complexity class for cubic time: :math:`\\mathcal{O}(n^3)`
    """

    @staticmethod
    def transform_n(n: np.ndarray) -> np.ndarray:
        return np.vstack((np.ones(len(n)), n * n * n)).T


class exponential(complexity):
    """
    Complexity class for exponential time: :math:`\\mathcal{O}(2^n)`
    """

    @staticmethod
    def transform_n(n: np.ndarray) -> np.ndarray:
        return np.vstack((np.ones(len(n)), n)).T

    @staticmethod
    def transform_t(t: np.ndarray) -> np.ndarray:
        return np.log2(t)


complexity_classes = [constant, logarithmic, linear, linearithmic, quadratic, cubic]#, exponential]



"""PyBryt plugin for Otter-Grader"""

import os
import dill
import shutil
import base64
import itertools
import warnings
import nbformat

from glob import glob
from otter.assign.assignment import Assignment
from otter.plugins import AbstractOtterPlugin
from otter.test_files import GradingResults
from otter.utils import get_source
from typing import Any, Dict, List, NoReturn, Optional, Union

from .. import ReferenceImplementation, StudentImplementation
from ..execution import NBFORMAT_VERSION
from ..utils import save_notebook


class OtterPlugin(AbstractOtterPlugin):
    """
    Plugin for Otter-Grader.
    """

    IMPORTABLE_NAME = "pybryt.integrations.otter.OtterPlugin"
    _cached_refs = None
    _generate_report = None
    _student_impl = None

    def during_assign(self, assignment: Assignment) -> NoReturn:
        """
        This event runs during ``otter assign`` and compiles the list of references indicated in the
        plugin config, placing the pickled reference implementations in the otuput directories. The
        stem of the filename for each reference is kept the same.

        Args:
            assignment (``otter.assign.assignment.Assignment``): the assignment configurations
        """
        self._cached_refs = []
        for fn in self.plugin_config["references"]:
            ref = ReferenceImplementation.compile(fn)
            if not isinstance(ref, list):
                ref = [ref]
            
            self._cached_refs.extend(ref)
            
            ref_fn = os.path.splitext(os.path.split(fn)[1])[0] + ".pkl"
            agp = assignment.result / 'autograder' / ref_fn
            with open(agp, "wb+") as f:
                dill.dump(ref, f)
            
            stp = assignment.result / 'student' / ref_fn
            shutil.copy(str(agp), str(stp))

    def during_generate(self, otter_config: Dict[str, Any], assignment: Assignment) -> NoReturn:
        """
        This event runs during ``otter generate`` and, if ``otter assign`` was run, adds the cached
        reference implementations from that run or compiles the indicated references if it was not
        into the ``otter_config`` as a base-64-encoded string.

        Args:
            otter_config (``dict``): the grading configurations
            assignment (``otter.assign.assignment.Assignment``): the assignment configurations; 
                should be set to ``None`` if ``otter assign`` was not run
        """
        if assignment is not None:
            cwd = os.getcwd()
            os.chdir(assignment.master.parent)

        cfg_idx = [self.IMPORTABLE_NAME in c.keys() if isinstance(c, dict) else False for c in otter_config["plugins"]].index(True)
        if self._cached_refs is None:
            refs = []
            for fn in otter_config["plugins"][cfg_idx][self.IMPORTABLE_NAME]["references"]:
                ref = ReferenceImplementation.compile(fn)
                if not isinstance(ref, list):
                    ref = [ref]
                refs.extend(ref)
        else:
            # if during_assign was run, this should be the same
            refs = self._cached_refs
        
        refs = base64.b64encode(dill.dumps(refs)).decode("ascii")
        otter_config["plugins"][cfg_idx][self.IMPORTABLE_NAME]["reference_bytes"] = refs

        if assignment is not None:
            os.chdir(cwd)

    def _generate_impl_report(self, refs: List[ReferenceImplementation], group: Optional[str] = None) -> NoReturn:
        """
        Generates and caches a student implementation from the submission path being graded as well
        as a report of what reference(s) were satisfied, if any, and the messages received from 
        those references.

        Args:
            refs (``list[ReferenceImplementation]``): the reference implementations to check against
            group (``str``, optional): a specific question group to run
        """
        nb = nbformat.read(self.submission_path, as_version=NBFORMAT_VERSION)
        self._remove_plugin_calls(nb)

        stu = StudentImplementation(nb)
        results = stu.check(refs, group=group)

        correct = [r.correct for r in results]
        
        report = f"PyBryt Reference Messages:\n"
        
        if any(correct):
            results = [results[correct.index(True)]]
            ending = "\nREFERENCE SATISFIED"

        else:
            ending = "\nNO REFERENCE SATISFIED"

        # TODO: separate mssages for separate references
        for m in itertools.chain(*(r.messages for r in results)):
            report += f"- {m}\n"
        
        report += ending

        self._generated_report = report
        self._student_impl = stu
    
    def _cache_student_impl(self, results: GradingResults, stu: StudentImplementation) -> NoReturn:
        """
        Caches the student implementation object as a base-64-encoded string in the grading results
        to be retrieved outside of the grading process.

        Args:
            results (``otter.test_files.GradingResults``): the grading results
            stu (``StudentImplementation``): the student implementation to cache
        """
        encoded_stu = stu.dumps()
        data = results.get_plugin_data(self.IMPORTABLE_NAME, {})
        data["cached_student_impl"] = encoded_stu
        results.set_plugin_data(self.IMPORTABLE_NAME, data)

    @classmethod
    def _remove_plugin_calls(cls, nb: Dict[str, Any]) -> NoReturn:
        """
        Removes calls to this Otter plugin from a notebook to ensure that a loop isn't created.
        Modifies the notebook in-place.

        Args:
            nb (``dict``): the notebook to remove calls from
        """
        for cell in nb['cells']:
            source = get_source(cell)
            for i in range(len(source)):
                if f".run_plugin(\"{cls.IMPORTABLE_NAME}\"" in source[i] or \
                        f".add_plugin_files(\"{cls.IMPORTABLE_NAME}\"" in source[i]:
                    source[i] = "# " + source[i]
            cell["source"] = "\n".join(source)

    def from_notebook(self, *ref_paths: str, group: Optional[str] = None) -> NoReturn:
        """
        This event runs when ``otter.Notebook.run_plugin`` is called to execute this plugin. It
        attempts to force-save the notebook and then loads the references indicated and generates
        an implementation report by running the references against the student's notebook. Prints
        the generated report to stdout.

        Args:
            *ref_paths (``str``): paths to reference implementation files
            group (``str``, optional): a specific question group to run
        """
        saved = save_notebook(self.submission_path)
        if not saved:
            warnings.warn(
                "Could not force-save notebook; the results of this call will be based on the last "
                "saved version of this notebook."
            )

        refs = []
        for rp in ref_paths:
            with open(rp, "rb") as f:
                ref = dill.load(f)
            # ref = ReferenceImplementation.load(rp)
            if isinstance(ref, ReferenceImplementation):
                refs.append(ref)
            else:
                refs.extend(ref)
        
        self._generate_impl_report(refs, group=group)
        print(self._generated_report)
    
    def notebook_export(self, dest: str = "student.pkl") -> List[str]:
        """
        Dumps the cached student implementation from a called to ``from_notebook`` into a file and
        returns the path to that file for exporting in the zip file.

        Args:
            dest (``str``, optional): the path at which to write the student implementation file
        """
        if self._student_impl is not None:
            self._student_impl.dump(dest)
        else:
            raise RuntimeError("Could not find a cached student implementation to export")
        return [dest]

    def before_execution(self, submission: Dict[str, Any]) -> dict:
        """
        Preprocessor for removing calls to the PyBryt plugin from a notebook before it is executed.

        Args:
            submission (``dict``): the submission notebook
        
        Returns:
            ``dict``: the modified submission notebook
        """
        self._remove_plugin_calls(submission)
        return submission

    def after_grading(self, results: GradingResults) -> NoReturn:
        """
        Generates an implementation report and caches the student implementation in the grading 
        results.

        Args:
            results (``otter.test_files.GradingResults``): the grading results
        """
        if self._student_impl is None:
            ref_bytes = base64.b64decode(self.plugin_config["reference_bytes"])
            refs = dill.loads(ref_bytes)
            self._generate_impl_report(refs)
        
        self._cache_student_impl(results, self._student_impl)

    def generate_report(self) -> str:
        """
        Returns the cached report, if present; otherwise, generates the implementation report and
        returns it.

        Returns:
            ``str``: the implementation report
        """
        if self._generated_report is None:
            ref_bytes = base64.b64decode(self.plugin_config["reference_bytes"])
            refs = dill.loads(ref_bytes)
            self._generate_impl_report(refs)
        
        return self._generated_report
    
    @classmethod
    def load_cached_implementations(cls, results: Union[GradingResults, List[GradingResults]]) -> \
            Union[StudentImplementation, List[StudentImplementation]]:
        """
        Loads any student implementations cached in the provided grading results.

        Args:
            results (``otter.test_files.GradingResults`` or ``list[otter.test_files.GradingResults]``):
                one or more grading results
        
        Returns:
            ``Union[StudentImplementation, List[StudentImplementation]]``: a student implementation,
            if a single result was passed, or a list thereof
        """
        if isinstance(results, GradingResults):
            data = results.get_plugin_data(cls.IMPORTABLE_NAME)["cached_student_impl"]
            return StudentImplementation.loads(data)
        elif isinstance(results, list):
            if not all(isinstance(r, GradingResults) for r in results):
                raise TypeError("Object of unexpected type received in results")
            data = [r.get_plugin_data(cls.IMPORTABLE_NAME) for r in results]
            return [StudentImplementation.loads(d["cached_student_impl"]) for d in data]
        else:
            raise TypeError("invalid type for argument 'results'")



""""""

from collections.abc import Sized
from contextlib import contextmanager
from typing import Union


_TRACKING_DISABLED = False


class TimeComplexityResult:
    """
    A class for collecting the results of time complexity check blocks.

    Args:
        name (``str``): the name of the block
        n (``int``): the input length
        start (``int``): the step counter value at the start of the block
        stop (``int``): the step counter value at the end of the block
    """

    name: str
    """the name of the block"""

    n: int
    """the input length or the input itself"""

    start: int
    """the step counter value at the start of the block"""

    stop: int
    """the step counter value at the end of the block"""

    def __init__(self, name, n, start, stop):
        self.name = name
        self.n = n
        self.start = start
        self.stop = stop


class check_time_complexity:
    """
    A context manager for checking the time complexity of student code.

    Halts tracking of values in memory and sets the trace function to only increment the step
    counter. When the block exits, the step counter is checked an a ``TimeComplexityResult`` object
    is appended to the student's memory footprint.

    If the current call stack is not being traced by PyBryt, no action is taken.

    Args:
        name (``str``): the name of the check; should match with the name of an annotation in the
            reference implementation
        n (``Union[int, float, Sized]``): the input length or the input itself if it supports ``len``
    """

    def __init__(self, name: str, n: Union[int, float, Sized]):
        if isinstance(n, float):
            n = int(n)
        if isinstance(n, Sized):
            n = len(n)
        if not isinstance(n, int):
            try:
                n = int(n)
            except:
                raise TypeError(f"n has invalid type {type(n)}")
  
        self._name = name
        self._n = n
        self._curr_steps = None
        self._observed, self._counter = None, None

    def __enter__(self):
        global _TRACKING_DISABLED

        from .tracing import _COLLECTOR_RET
        if _COLLECTOR_RET is not None:
            self._observed, self._counter, _ = _COLLECTOR_RET
            self._curr_steps = self._counter[0]

        _TRACKING_DISABLED = True

    def __exit__(self, exc_type, exc_value, traceback):
        global _TRACKING_DISABLED

        _TRACKING_DISABLED = False

        if self._curr_steps is not None:
            end_steps = self._counter[0]
            self._observed.append((
                TimeComplexityResult(self._name, self._n, self._curr_steps, end_steps), 
                end_steps,
            ))

        return False



"""Command-line interface for PyBryt"""

import os
import dill
import json
import click

from . import (
    generate_report, generate_student_impls, ReferenceImplementation, StudentImplementation, __version__
)
from .utils import get_stem


@click.group()
@click.version_option(__version__)
def click_cli():
    """
    A command-line interface for PyBryt. See https://microsoft.github.io/pybryt for more information.
    """
    pass


@click_cli.command("compile")
@click.option("-d", "--dest", default=None, type=click.Path(dir_okay=False), 
              help="Path at which to write the pickled reference implementation")
@click.option("-n", "--name", default=None, type=click.STRING, 
              help="Optional name for the reference implementation")
@click.argument("src", type=click.Path(exists=True, dir_okay=False))
def compile_reference(src, dest, name):
    """
    Compile the reference implementation SRC.

    SRC must be a path to a Jupyter Notebook file which defines a reference implementation.
    """
    ref = ReferenceImplementation.compile(src, name=name)
    ref.dump(dest)


@click_cli.command()
@click.option("-n", "--name", default=None, type=click.STRING, 
              help="Optional name for the reference implementation")
@click.option("--output-nb", default=None, type=click.Path(dir_okay=False), 
              help="Path at which to write the output notebook from executing the student submission")
@click.option("-o", "--output", default=None, type=click.Path(dir_okay=False), 
              help="Path at which to write the results of the check")
@click.option("-t", "--type", "output_type", default="pickle", show_default=True, 
              type=click.Choice(["pickle", "json", "report"]), help="Type of output to write")
@click.argument("ref", type=click.Path(exists=True, dir_okay=False))
@click.argument("stu", type=click.Path(exists=True, dir_okay=False))
def check(ref, stu, name, output_nb, output, output_type):
    """
    Run a student submission against a reference implementation.

    REF can be a path to a pickled reference implementation or to a notebook to be compiled on-the-fly.
    STU can be a path to a pickled student implementation or to a notebook to be executed.

    If TYPE is "pickle" or "json", the output is a file. If TYPE is "report", a report is echoed to
    the console and OUTPUT is ignored.
    """
    if os.path.splitext(ref)[1] == ".ipynb":
        ref = ReferenceImplementation.compile(ref)
    else:
        try:
            ref = ReferenceImplementation.load(ref)
        except:
            raise RuntimeError(f"Could not load the reference implementation {ref}")

    if output is None:
        output = get_stem(stu) + "_results" + (".pkl", ".json")[output_type == "json"]

    if os.path.splitext(stu)[1] == ".ipynb":
        stu = StudentImplementation(stu, output=output_nb)
    else:
        try:
            stu = StudentImplementation.load(stu)
        except:
            raise RuntimeError(f"Could not load the student implementation {stu}")

    res = stu.check(ref)

    if output_type == "pickle":
        if isinstance(res, list):
            with open(output, "wb+") as f:
                dill.dump(res, f)
        else:
            res.dump(output)
    elif output_type == "json":
        if isinstance(res, list):
            d = [r.to_dict() for r in res]
        else:
            d = res.to_dict()
        with open(output, "w+") as f:
            json.dump(d, f, indent=2)
    elif output_type == "report":
        report = generate_report(res)
        click.echo(report)


@click_cli.command()
@click.option("-p", "--parallel", is_flag=True, default=False, 
              help="Execute notebooks in parallel using the multiprocessing library")
@click.option("-o", "--output", default=None, type=click.Path(), 
              help="Path at which to write the pickled student implementation")
@click.argument("subm", nargs=-1, type=click.Path(exists=True, dir_okay=False))
def execute(subm, parallel, output):
    """
    Execute student submissions to generate memory footprints.

    Executes the student submission(s) SUBM at writes the pickled objects to some output file. If
    OUTPUT is unspecified, this defaults to "./{SUBM.stem}.pkl" (e.g. for SUBM 
    "submissions/subm01.ipynb", this is "./subm01.pkl").
    """
    if len(subm) == 0:
        raise ValueError("You must specify at least one notebook to execute")

    stus = generate_student_impls(subm, parallel=parallel)

    if len(subm) == 1:
        stus[0].dump(output)

    else:
        if output is None:
            output = "./"

        if not os.path.isdir(output):
            raise ValueError(f"Output directory {output} does not exist or is not a directory")

        for s, stu in zip(subm, stus):
            stem = get_stem(s)
            path = os.path.join(output, stem + ".pkl")
            stu.dump(path)


def cli(*args, **kwargs):
    """
    Wrapper for the click CLI that sets the prog name.
    """
    prog_name = kwargs.pop("prog_name", "pybryt")
    return click_cli(*args, prog_name=prog_name, **kwargs)



"""Plagiarism checking for PyBryt"""

import random
import numpy as np

from typing import List, Union

from .annotations import Value
from .reference import ReferenceImplementation, ReferenceResult


def create_references(student_impls: List["StudentImplementation"], frac=0.25, seed=None, 
        filter_none=True, **kwargs) -> List[ReferenceImplementation]:
    """
    """
    if seed is not None:
        random.seed(seed)

    values = [[t[0] for t in stu.values if t[0] is not None] for stu in student_impls]

    k = int(min([len(v) for v in values]) * frac)
    
    refs = []
    for stu, vals in zip(student_impls, values):
        ref_values = random.sample(vals, k)
        refs.append(ReferenceImplementation([Value(v) for v in ref_values]))

    return refs


def get_impl_results(ref_impl: ReferenceImplementation, student_impls: List["StudentImplementation"], 
        arr=True, **kwargs) -> Union[List[ReferenceResult], np.ndarray]:
    """
    Returns matrix where rows are student impls and cols represent whether each value was satisfied
    by the student impl
    """
    # results = []
    # for ref in ref_impls:
    #     ref_results = []
    #     for stu in student_impls:
    #         ref_results.append(stu.check(ref))
    #     results.append(ref_results)

    results = []
    for stu in student_impls:
        results.append(stu.check(ref_impl))
    
    if arr:
        return np.array([r.to_array() for r in results])
    else:
        return results


def compare_implementations(student_impls: List["StudentImplementation"], **kwargs) -> \
        List[Union[List[ReferenceResult], np.ndarray]]:
    """
    """
    refs = create_references(student_impls, **kwargs)
    results = []
    for ref in refs:
        results.append(get_impl_results(ref, student_impls, **kwargs))
    return results


from .student import StudentImplementation



""""""

import pytest
import nbformat
import tempfile

from click.testing import CliRunner
from unittest import mock

from pybryt import StudentImplementation
from pybryt.cli import click_cli
from pybryt.utils import get_stem

from .test_reference import generate_reference_notebook
from .test_student import generate_impl


def test_check():
    """
    """
    runner = CliRunner()
    with mock.patch("pybryt.cli.ReferenceImplementation") as mocked_ref, \
            mock.patch("pybryt.cli.StudentImplementation") as mocked_stu, \
            mock.patch("pybryt.cli.dill") as mocked_dill, \
            mock.patch("pybryt.cli.json") as mocked_json, \
            mock.patch("pybryt.cli.open") as mocked_open:
        mocked_stu.return_value = mock.create_autospec(StudentImplementation)

        with tempfile.NamedTemporaryFile(mode="w+", suffix=".ipynb") as ref_ntf, \
                tempfile.NamedTemporaryFile(mode="w+", suffix=".ipynb") as stu_ntf:
            result = runner.invoke(click_cli, ["check", ref_ntf.name, stu_ntf.name])
            assert result.exit_code == 0
            mocked_ref.compile.assert_called_with(ref_ntf.name)
            mocked_stu.assert_called_with(stu_ntf.name, output=None)
            mocked_stu.return_value.check.return_value.dump.assert_called_with(get_stem(stu_ntf.name) + "_results.pkl")

        with tempfile.NamedTemporaryFile(mode="w+", suffix=".pkl") as ref_ntf, \
                tempfile.NamedTemporaryFile(mode="w+", suffix=".pkl") as stu_ntf:
            result = runner.invoke(click_cli, ["check", ref_ntf.name, stu_ntf.name, "-t", "json"])
            assert result.exit_code == 0
            mocked_ref.load.assert_called_with(ref_ntf.name)
            mocked_stu.load.assert_called_with(stu_ntf.name)
            mocked_stu.load.return_value.check.return_value.to_dict.assert_called_with()
            mocked_json.dump.assert_called()

            with mock.patch("pybryt.cli.generate_report") as mocked_report:
                result = runner.invoke(click_cli, ["check", ref_ntf.name, stu_ntf.name, "-t", "report"])
                assert result.exit_code == 0
                mocked_ref.load.assert_called_with(ref_ntf.name)
                mocked_stu.load.assert_called_with(stu_ntf.name)
                mocked_report.assert_called_with(mocked_stu.load.return_value.check.return_value)

            # check with list of references
            mocked_ref.load.return_value = []
            mocked_stu.load.return_value.check.return_value = []
            
            result = runner.invoke(click_cli, ["check", ref_ntf.name, stu_ntf.name])
            assert result.exit_code == 0
            mocked_ref.load.assert_called_with(ref_ntf.name)
            mocked_stu.load.assert_called_with(stu_ntf.name)
            mocked_dill.dump.assert_called()

            result = runner.invoke(click_cli, ["check", ref_ntf.name, stu_ntf.name, "-t", "json"])
            assert result.exit_code == 0
            mocked_ref.load.assert_called_with(ref_ntf.name)
            mocked_stu.load.assert_called_with(stu_ntf.name)
            mocked_json.dump.assert_called()

            # check errors
            mocked_stu.load.side_effect = Exception()
            result = runner.invoke(click_cli, ["check", ref_ntf.name, stu_ntf.name])
            assert result.exit_code == 1
            assert isinstance(result.exception, RuntimeError)
            assert result.exception.args[0] == f"Could not load the student implementation {stu_ntf.name}"

            mocked_ref.load.side_effect = Exception()
            result = runner.invoke(click_cli, ["check", ref_ntf.name, stu_ntf.name])
            assert result.exit_code == 1
            assert isinstance(result.exception, RuntimeError)
            assert result.exception.args[0] == f"Could not load the reference implementation {ref_ntf.name}"


def test_compile():
    """
    """
    runner = CliRunner()
    with tempfile.NamedTemporaryFile(mode="w+", suffix=".ipynb") as ntf:
        nb = generate_reference_notebook()
        nbformat.write(nb, ntf)
        ntf.seek(0)

        with mock.patch("pybryt.cli.ReferenceImplementation") as mocked_ref:
            result = runner.invoke(click_cli, ["compile", ntf.name])
            assert result.exit_code == 0
            mocked_ref.compile.assert_called_with(ntf.name, name=None)
            mocked_ref.compile.return_value.dump.assert_called_with(None)

            result = runner.invoke(click_cli, ["compile", ntf.name, "-n", "foo"])
            assert result.exit_code == 0
            mocked_ref.compile.assert_called_with(ntf.name, name="foo")
            mocked_ref.compile.return_value.dump.assert_called_with(None)

            result = runner.invoke(click_cli, ["compile", ntf.name, "-d", "foo.pkl"])
            assert result.exit_code == 0
            mocked_ref.compile.assert_called_with(ntf.name, name=None)
            mocked_ref.compile.return_value.dump.assert_called_with("foo.pkl")


def test_execute():
    """
    """
    runner = CliRunner()
    num_subms = 10
    ntfs = []
    nb = generate_reference_notebook()
    for _ in range(num_subms):
        ntf = tempfile.NamedTemporaryFile(mode="w+", suffix=".ipynb")
        nbformat.write(nb, ntf)
        ntf.seek(0)
        ntfs.append(ntf)

    fns = tuple(ntf.name for ntf in ntfs)

    with mock.patch("pybryt.cli.generate_student_impls") as mocked_generate:
        mocked_generate.return_value.__iter__.return_value = (
            iter([mock.create_autospec(StudentImplementation) for _ in range(num_subms)])
        )

        result = runner.invoke(click_cli, ["execute", *fns])
        assert result.exit_code == 0
        mocked_generate.assert_called_with(fns, parallel=False)

        result = runner.invoke(click_cli, ["execute", fns[0]])
        assert result.exit_code == 0
        mocked_generate.assert_called_with((fns[0], ), parallel=False)

        result = runner.invoke(click_cli, ["execute", "-p", *fns])
        assert result.exit_code == 0
        mocked_generate.assert_called_with(fns, parallel=True)

        # check for error on nonexistance output dir
        result = runner.invoke(click_cli, ["execute", *fns, "-o", "/some/fake/path"])
        assert result.exit_code == 1
        assert isinstance(result.exception, ValueError)
        assert result.exception.args[0] == "Output directory /some/fake/path does not exist or is not a directory"

    # check other errors
    result = runner.invoke(click_cli, ["execute"])
    assert result.exit_code == 1
    assert isinstance(result.exception, ValueError)
    assert result.exception.args[0] == "You must specify at least one notebook to execute"


def test_cli_func(capsys):
    """
    Checks that the prog name is set correctly.
    """
    from pybryt.__main__ import cli
    try:
        cli(["--help"])
    except SystemExit:
        pass
    captured = capsys.readouterr()
    assert captured.out.startswith("Usage: pybryt [")



"""Annotations for checking type presence, or lack thereof, within memory footprints"""

__all__ = ["ForbidType"]

import dill

from typing import Any, Dict, List, Tuple

from .annotation import Annotation, AnnotationResult


class ForbidType(Annotation):
    """
    Annotation class for asserting there are no objects of a specific type in the memory footprint.

    Indicates that the student's memory footprint should not have any values of a certain type. Uses
    the ``isinstance`` function to determine if any values match the specified type. The type itself
    must be pickleable by ``dill``.

    Args:
        type\\_ (``type``): the type to forbid
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.annotations.annotation.Annotation>` constructor
    """

    type_: type
    """the type to forbid"""

    def __init__(self, type_, **kwargs):
        if not isinstance(type_, type):
            raise TypeError(f"{type_} is not a type")

        try:
            dill.dumps(type_)
        except Exception as e:
            raise ValueError(f"Types must be serializable but the following error was thrown during serialization:\n{e}")

        self.type_ = type_

        super().__init__(**kwargs)

    @property
    def children(self) -> List[Annotation]:
        return []

    def check(self, observed_values: List[Tuple[Any, int]]) -> AnnotationResult:
        """
        Checks that there are no values of type ``self.type_`` in the memory footprint.

        Args:
            observed_values (``list[tuple[object, int]]``): a list of tuples of values observed
                during execution and the timestamps of those values
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation based on ``observed_values``
        """
        for val, _ in observed_values:
            if isinstance(val, self.type_):
                return AnnotationResult(False, self)
        return AnnotationResult(True, self)

    def __eq__(self, other: Any) -> bool:
        """
        Checks whether this annotation is equal to another object.

        To be equal to a ``ForbidType``, the other object must also be a ``Forbidtype`` object, have
        the same instance variables, and the same forbidden type.

        Args:
            other (``object``): the object to compare to

        Returns:
            ``bool``: whether the objects are equal
        """
        return super().__eq__(other) and self.type_ == other.type_

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts this annotation's details to a JSON-friendly dictionary format.

        Output dictionary contains the annotation's name, group, limit number, success message, and
        failure message, as well as the type as a string.

        Returns:
            ``dict[str, object]``: the dictionary representation of this annotation
        """
        d = super().to_dict()
        d.update({
            "type_": str(self.type_),
        })
        return d



""""""

import os
import dill
import json
import base64
import tempfile
import nbformat
import numpy as np
import pkg_resources
import pytest

from copy import deepcopy
from functools import lru_cache
from textwrap import dedent
from unittest import mock

from pybryt import generate_report, ReferenceImplementation, Value
from pybryt.execution import execute_notebook


def generate_mfp(nb):
    return execute_notebook(nb, "")[1]


def generate_reference_notebook():
    """
    """
    nb = nbformat.v4.new_notebook()
    nb.cells.append(nbformat.v4.new_code_cell(dedent("""\
        import pybryt
    """)))
    nb.cells.append(nbformat.v4.new_code_cell(dedent("""\
        def median(S):
            sorted_S = sorted(S) 
            pybryt.Value(sorted_S, name="sorted", group="median", limit=5, success_message="SUCCESS: Sorted the sample correctly", 
                        failure_message="ERROR: The sample was not sorted")
            
            size_of_set = len(S) 
            pybryt.Value(size_of_set, name="size", group="median", success_message = "SUCCESS: Computed the size of the sample", 
                        failure_message="ERROR: Did not capture the size of the set to determine if it is odd or even")
            
            middle = size_of_set // 2
            is_set_size_even = (size_of_set % 2) == 0

            if is_set_size_even:
                return (sorted_S[middle-1] + sorted_S[middle]) / 2
            else:
                return sorted_S[middle]
    """)))
    nb.cells.append(nbformat.v4.new_code_cell(dedent("""\
        import numpy as np
        np.random.seed(42)
        for _ in range(10):
            vals = [np.random.randint(-1000, 1000) for _ in range(np.random.randint(1, 1000))]
            val = median(vals)
            pybryt.Value(val, name="median", group="median", success_message="SUCCESS: computed the correct median", 
                failure_message="ERROR: failed to compute the median")
    """)))
    return nb


def test_reference_construction():
    """
    """
    nb = generate_reference_notebook()

    # check compile error
    with pytest.raises(ValueError, match="No name specified for the reference being compiled"):
        ReferenceImplementation.compile(nb)

    ref = ReferenceImplementation.compile(nb, name="foo")

    # test ReferenceImplementation.get
    sorted_annots = ref.get("sorted")
    assert len(sorted_annots) == 5
    assert all(isinstance(a, Value) for a in sorted_annots)

    with pytest.raises(ValueError, match="Found no annotations with name 'foo'"):
        ref.get("foo")

    ref_filename = pkg_resources.resource_filename(__name__, os.path.join("files", "expected_ref.pkl"))
    expected_ref = ReferenceImplementation.load(ref_filename)

    with tempfile.NamedTemporaryFile() as ntf:
        ref.dump(ntf.name)
        second_ref = ReferenceImplementation.load(ntf.name)
        assert ref == second_ref
        assert ref == expected_ref

    # test construction from .py file w/ ReferenceImplementation objects
    ref2_filename = pkg_resources.resource_filename(__name__, os.path.join("files", "expected_ref2.pkl"))
    expected_ref2 = ReferenceImplementation.load(ref2_filename)

    with tempfile.NamedTemporaryFile("w+", suffix=".py") as ntf:
        ntf.write(dedent("""\
            import pybryt
            import numpy as np
            np.random.seed(42)

            def square_evens(arr):
                subarr = arr[arr % 2 == 0]
                v1 = pybryt.Value(subarr)
                subarr = subarr ** 2
                v2 = pybryt.Value(subarr)
                arr = arr.copy()
                arr[arr % 2 == 0] = subarr
                return v1, v2, arr

            annots = []
            for _ in range(10):
                vals = np.array([np.random.randint(-1000, 1000) for _ in range(np.random.randint(1, 1000))])
                v1, v2, val = square_evens(vals)
                annots.append(v1)
                annots.append(v2)
                annots.append(pybryt.Value(val))
            
            ref = pybryt.ReferenceImplementation("foo", annots)
            ref2 = pybryt.ReferenceImplementation("bar", [])
        """))

        ntf.seek(0)

        more_refs = ReferenceImplementation.compile(ntf.name, name="foo")
        assert len(more_refs) == 2
        assert len(more_refs[1].annotations) == 0
        
        ref2 = more_refs[0]
        assert ref2 == expected_ref2


def test_construction_errors():
    """
    """
    with pytest.raises(TypeError, match="annotations should be a list of Annotations"):
        ReferenceImplementation("foo", set())

    with pytest.raises(TypeError, match="Found non-annotation in annotations"):
        ReferenceImplementation("bar", [Value(1), Value(2), 3, Value(4)])

    # check that you can't load something that isn't a ReferenceImplementation
    with tempfile.NamedTemporaryFile() as ntf:
        dill.dump(1, ntf)

        ntf.seek(0)

        with pytest.raises(TypeError, match="Unpickled object is not of type <class 'pybryt.reference.ReferenceImplementation'>"):
            ReferenceImplementation.load(ntf.name)

    # check that loading an empty reference implementation gives an error
    with tempfile.NamedTemporaryFile(mode="w+", suffix=".ipynb") as ntf:
        nbformat.write(nbformat.v4.new_notebook(), ntf)

        ntf.seek(0)

        with pytest.warns(UserWarning, match=f"Could not find any reference implementations in {ntf.name}"):
            ReferenceImplementation.compile(ntf.name, name="bar")


def test_run_and_results():
    """
    """
    nb = generate_reference_notebook()
    nb.cells.append(nbformat.v4.new_code_cell(dedent("""\
        vals = [np.random.randint(-1000, 1000) for _ in range(np.random.randint(1, 1000))]
        val = median(vals)
        # this annotation is not in the 'median' group
        pybryt.Value(val, success_message="SUCCESS: computed the correct median x2", 
            failure_message="ERROR: failed to compute the median")
    """)))
    ref = ReferenceImplementation.compile(nb, name="foo")
    # _, vals = execute_notebook(nb, "")
    vals = generate_mfp(nb)
    
    res = ref.run(vals)
    assert res.name == ref.name
    assert len(res.results) == 27
    assert res.reference is ref
    assert res.correct is True
    assert (res.to_array() == np.ones(27)).all()
    assert repr(res).startswith("ReferenceResult([\n") and len(repr(res).split("\n")) == 29
    assert res.messages == [
        'SUCCESS: Sorted the sample correctly', 
        'SUCCESS: Computed the size of the sample', 
        'SUCCESS: computed the correct median',
        'SUCCESS: computed the correct median x2',
    ]

    res = ref.run(vals, group="median")
    assert len(res.results) == 26
    assert res.reference is ref
    assert res.correct is True
    assert (res.to_array() == np.ones(26)).all()
    assert repr(res).startswith("ReferenceResult([\n") and len(repr(res).split("\n")) == 28
    assert res.messages == [
        'SUCCESS: Sorted the sample correctly', 
        'SUCCESS: Computed the size of the sample', 
        'SUCCESS: computed the correct median',
    ]

    nb.cells.insert(2, nbformat.v4.new_code_cell("import numpy as np\ndef median(S):\n    return np.median(S)"))
    # _, vals = execute_notebook(nb, "")
    vals = generate_mfp(nb)
    
    res = ref.run(vals)
    assert len(res.results) == 27
    assert res.reference is ref
    assert res.correct is False
    assert (res.to_array() == np.array([0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 
        1, 1, 1, 1, 1, 1, 1])).all()
    assert repr(res).startswith("ReferenceResult([\n") and len(repr(res).split("\n")) == 29
    assert res.messages == [
        'ERROR: The sample was not sorted',
        'ERROR: Did not capture the size of the set to determine if it is odd or even',
        'SUCCESS: computed the correct median',
        'SUCCESS: computed the correct median x2',
    ]

    res_filename = pkg_resources.resource_filename(__name__, os.path.join("files", "expected_result.json"))
    with open(res_filename) as f:
        expected_res_dict = json.load(f)

    for d in expected_res_dict["results"]:
        d.pop("satisfied_at")

    res_dict = res.to_dict()

    # check satisfied_at field here since they won't match up
    assert all(isinstance(d.pop("satisfied_at", None), int) for d in res_dict["results"])

    assert res_dict == expected_res_dict

    with pytest.raises(ValueError, match="Group 'foo' not found"):
        ref.run(vals, group="foo")


def test_generate_report():
    """
    """
    nb = generate_reference_notebook()
    ref = ReferenceImplementation.compile(nb, name="foo")
    # _, vals = execute_notebook(nb, "")
    vals = generate_mfp(nb)
    res = ref.run(vals)

    report = generate_report(res)
    assert report == dedent("""\
        REFERENCE: foo
        SATISFIED: True
        MESSAGES:
          - SUCCESS: Sorted the sample correctly
          - SUCCESS: Computed the size of the sample
          - SUCCESS: computed the correct median
    """).strip()

    res2 = deepcopy(res)
    res2.results[0]._satisfied = False
    report = generate_report(res2)
    assert report == dedent("""\
        REFERENCE: foo
        SATISFIED: False
        MESSAGES:
          - ERROR: The sample was not sorted
          - SUCCESS: Computed the size of the sample
          - SUCCESS: computed the correct median
    """).strip()

    report = generate_report(res, show_only="unsatisfied")
    assert report == dedent("""\
        REFERENCE: foo
        SATISFIED: True
        MESSAGES:
          - SUCCESS: Sorted the sample correctly
          - SUCCESS: Computed the size of the sample
          - SUCCESS: computed the correct median
    """).strip()

    report = generate_report(res, show_only="unsatisfied", fill_empty=False)
    assert report == ""

    res = [res, res2]
    report = generate_report(res)
    assert report == dedent("""\
        REFERENCE: foo
        SATISFIED: True
        MESSAGES:
          - SUCCESS: Sorted the sample correctly
          - SUCCESS: Computed the size of the sample
          - SUCCESS: computed the correct median

        REFERENCE: foo
        SATISFIED: False
        MESSAGES:
          - ERROR: The sample was not sorted
          - SUCCESS: Computed the size of the sample
          - SUCCESS: computed the correct median
    """).strip()

    report = generate_report(res, show_only="satisfied")
    assert report == dedent("""\
        REFERENCE: foo
        SATISFIED: True
        MESSAGES:
          - SUCCESS: Sorted the sample correctly
          - SUCCESS: Computed the size of the sample
          - SUCCESS: computed the correct median
    """).strip()

    report = generate_report(res, show_only="unsatisfied")
    assert report == dedent("""\
        REFERENCE: foo
        SATISFIED: False
        MESSAGES:
          - ERROR: The sample was not sorted
          - SUCCESS: Computed the size of the sample
          - SUCCESS: computed the correct median
    """).strip()

    # test misc. errors
    with pytest.raises(TypeError, match="Cannot generate a report from arguments that are not reference result objects"):
        generate_report(1)

    with pytest.raises(TypeError, match="Cannot generate a report from arguments that are not reference result objects"):
        generate_report([1])
    
    with pytest.raises(ValueError, match="show_only must be in {'satisfied', 'unsatisfied', None}"):
        generate_report(res, show_only="foo")



"""Abstract base class for annotations and annotation results class"""

__all__ = ["Annotation", "AnnotationResult"]

from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Any, Dict, List, NoReturn, Optional, Tuple


_TRACKED_ANNOTATIONS = []
_GROUP_INDICES = {}
_ANNOTATION_COUNTER = 0


class Annotation(ABC):
    """
    Abstract base class for annotating a reference implementation.

    Defines an API and global configurations for all annotations used for marking-up reference
    implementations. All instances of this class, or any of its subclasses, are added to a singleton
    list that PyBryt maintains to track all annotations for simple reference creation. Supports
    bitwise logical operators and contains methods for creating relational annotations (see
    :py:class:`RelationalAnnotation<pybryt.RelationalAnnotation>`).

    Args:
        name (``str``, optional): the name of the annotation
        limit (``int``, optional): the maximum number of annotations with name ``name`` to track
            in ``_TRACKED_ANNOTATIONS``
        group (``str``, optional): the name of the group that this annotation belongs to
        success_message (``str``, optional): a message to relay to the student if satisfied
        failure_message (``str``, optional): a message to relay to the student if not satisfied
    """

    name: str
    """the name of the annotation"""
    
    limit: Optional[int]
    """the maximum number of annotations with this name to track in :obj:`_TRACKED_ANNOTATIONS`"""
    
    group: Optional[str]
    """the name of the group that this annotation belongs to"""
    
    success_message: Optional[str]
    """a message to relay to the student if satisfied"""
    
    failure_message: Optional[str]
    """a message to relay to the student if not satisfied"""

    def __init__(
        self, name: Optional[str] = None, limit: Optional[int] = None, group: Optional[str] = None, 
        success_message: Optional[str] = None, failure_message: Optional[str] = None,
    ):
        global _ANNOTATION_COUNTER
        _ANNOTATION_COUNTER += 1
        if name is not None:
            self.name = name
        else:
            self.name = f"Annotation {_ANNOTATION_COUNTER}"
        self.limit = limit
        self.group = group
        self.success_message = success_message
        self.failure_message = failure_message
        
        self._track()

    def __repr__(self):
        ret = f"pybryt.{self.__class__.__name__}"
        return ret
    
    def _track(self) -> NoReturn:
        """
        Tracks this annotation in ``_TRACKED_ANNOTATIONS`` and updates ``_GROUP_INDICES`` with the
        index of the annotation if ``self.group`` is present. If the annotation has children
        (returned by ``self.children``), the children are removed from ``_TRACKED_ANNOTATIONS``.
        """
        from ..execution.complexity import _TRACKING_DISABLED
        if _TRACKING_DISABLED:
            return

        global _GROUP_INDICES, _TRACKED_ANNOTATIONS

        idx = len(_TRACKED_ANNOTATIONS)
        if self.name not in _GROUP_INDICES:
            _GROUP_INDICES[self.name] = []
        if self.limit is not None and len(_GROUP_INDICES[self.name]) >= self.limit:
            return
        else:
            _GROUP_INDICES[self.name].append(idx)
    
        for child in self.children:
            try:
                _TRACKED_ANNOTATIONS.remove(child)
            except ValueError:
                pass
        
        _TRACKED_ANNOTATIONS.append(self)
    
    @staticmethod
    def get_tracked_annotations() -> List["Annotation"]:
        """
        Returns the list of tracked annotations.

        Returns:
            ``list[Annotation]``: the list of tracked annotations
        """
        return _TRACKED_ANNOTATIONS

    @staticmethod
    def reset_tracked_annotations() -> NoReturn:
        """
        Resets the list of tracked annotations and the mapping of group names to indices in that
        list.
        """
        global _ANNOTATION_COUNTER, _GROUP_INDICES, _TRACKED_ANNOTATIONS
        _TRACKED_ANNOTATIONS.clear()
        _GROUP_INDICES.clear()
        _ANNOTATION_COUNTER = 0

    @property
    @abstractmethod
    def children(self) -> List["Annotation"]:
        """
        ``list[Annotation]``: the child annotations of this annotation. If this annotation has no 
        children, an empty list.
        """
        ... # pragma: no cover

    @abstractmethod
    def check(self, observed_values: List[Tuple[Any, int]]) -> "AnnotationResult":
        """
        Runs the check on the condition asserted by this annotation and returns a results object.

        Checks that the condition required by this annotation is met using the list of tuples of
        observed values and timestamps ``observed_values``. Creates and returns an 
        :py:class:`AnnotationResult<pybryt.AnnotationResult>` object with the results of this check.

        Args:
            observed_values (``list[tuple[object, int]]``): a list of tuples of values observed
                during execution and the timestamps of those values
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation based on 
            ``observed_values``
        """
        ... # pragma: no cover

    def __eq__(self, other: Any) -> bool:
        """
        Checks whether this annotation is equal to another object.

        Args:
            other (``object``): the object to compare to

        Returns:
            ``bool``: whether the objects are equal
        """
        return isinstance(other, type(self)) and  self.name == other.name and \
            self.success_message == other.success_message and \
            self.failure_message == other.failure_message and self.group == other.group and \
            self.limit == other.limit

    def before(self, other_annotation: "Annotation", **kwargs) -> "BeforeAnnotation":
        """
        Creates an annotation asserting that this annotation is satisfied before another (i.e. that
        the satisfying timestamp of this annotation is less than or equal to that of 
        ``other_annotation``).

        Args:
            other_annotation (``Annotation``): the annotation that should be satisfied after this 
                one
            kwargs: other keyword arguments passed to the ``BeforeAnnotation`` constructor
        
        Returns:
            ``BeforeAnnotation``: the annotation asserting the before condition
        """
        return BeforeAnnotation(self, other_annotation, **kwargs)

    def after(self, other_annotation: "Annotation", **kwargs) -> "BeforeAnnotation":
        """
        Creates an annotation asserting that this annotation is satisfied after another (i.e. that
        the satisfying timestamp of this annotation is greater than or equal to that of 
        ``other_annotation``).

        Args:
            other_annotation (``Annotation``): the annotation that should be satisfied before this 
                one
            kwargs: other keyword arguments passed to the ``BeforeAnnotation`` constructor
        
        Returns:
            ``BeforeAnnotation``: the annotation asserting the before condition
        """
        return BeforeAnnotation(other_annotation, self, **kwargs)

    def __and__(self, other_annotation: "Annotation") -> "AndAnnotation":
        """
        Overrides the ``&`` operator for annotations to return an 
        :py:class`AndAnnotation<pybryt.AndAnnotation>`, asserting that both annotations should be 
        satisfied.

        Args:
            other_annotation (``Annotation``): the other annotation

        Returns:
            ``AndAnnotation``: the annotation asserting the and condition
        """
        return AndAnnotation(self, other_annotation)

    def __or__(self, other_annotation: "Annotation") -> "OrAnnotation":
        """
        Overrides the ``|`` operator for annotations to return an 
        :py:class`OrAnnotation<pybryt.OrAnnotation>`, asserting that either annotation should be 
        satisfied.

        Args:
            other_annotation (``Annotation``): the other annotation

        Returns:
            ``OrAnnotation``: the annotation asserting the and condition
        """
        return OrAnnotation(self, other_annotation)

    def __xor__(self, other_annotation: "Annotation") -> "XorAnnotation":
        """
        Overrides the ``^`` operator for annotations to return an 
        :py:class`XorAnnotation<pybryt.XorAnnotation>`, asserting that either annotation but not 
        both should be satisfied.

        Args:
            other_annotation (``Annotation``): the other annotation

        Returns:
            ``XorAnnotation``: the annotation asserting the xor condition
        """
        return XorAnnotation(self, other_annotation)

    def __invert__(self) -> "NotAnnotation":
        """
        Overrides the ``~`` operator for annotations to return an 
        :py:class`NotAnnotation<pybryt.NotAnnotation>`, asserting that this annotation should *not*
        be satisfied.

        Args:
            other_annotation (``Annotation``): the other annotation

        Returns:
            ``NotAnnotation``: the annotation asserting the and condition
        """
        return NotAnnotation(self)

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts this annotation's details to a JSON-friendly dictionary format.

        Output dictionary contains the annotation's name, group, limit number, success message, and
        failure message.

        Returns:
            ``dict[str, object]``: the dictionary representation of this annotation
        """
        type_name = type(self).__name__
        if type_name.endswith("Annotation"):
            type_name = type_name[:-len("Annotation")]
        type_name = type_name.lower()
        if type_name.startswith("_"):
            type_name = None
        return {
            "name": self.name,
            "group": self.group,
            "limit": self.limit,
            "success_message": self.success_message,
            "failure_message": self.failure_message,
            "children": [c.to_dict() for c in self.children],
            "type": type_name,
        }


class AnnotationResult:
    """
    Class that manages and defines an API for interacting with the results of an annotation.

    Created when an annotation calls its :py:meth:`check<pybryt.Annotation.check` method and 
    wrangles the results of that annotation. Contains fields for tracking child annotation results,
    values satisfying annotations, and messages returned by this annotation and its children.

    Args:
        satisfied (``bool`` or ``None``): whether the condition of the annotation was satisfied; if
            child annotation results should be used to determine this value, set to ``None``
        annotation (:py:class:`Annotation`): the annotation that this result is for
        value (``object``, optional): the value that satisfied the condition of this annotation
        timestamp (``int``, optional): the step counter value at which this annotation was satisfied
        children (``list[AnnotationResult]``, optional): child annotation results of this annotation
            result
    """

    _satisfied: Optional[bool]
    """
    whether the condition of the annotation was satisfied; if child annotation results should be 
    used to determine this value, set to ``None``
    """

    annotation: Annotation
    """the annotation that this result is for"""

    _value: Any
    """the value that satisfied the condition of this annotation"""

    timestamp: int
    """the step counter value at which this annotation was satisfied"""

    children: List["AnnotationResult"]
    """child annotation results of this annotation result"""


    def __init__(
        self, satisfied: Optional[bool], annotation: Annotation, value: Any = None, timestamp: int = -1, 
        children: List["AnnotationResult"] = [],
    ):
        self._satisfied = satisfied
        self.annotation = annotation
        self._value = value
        self.timestamp = timestamp
        self.children = children

    def __repr__(self):
        return f"AnnotationResult(satisfied={self.satisfied}, annotation={self.annotation})"

    @property
    def satisfied(self) -> bool:
        """
        ``bool``: whether this annotation was satisfied
        """
        if self._satisfied is not None:
            return self._satisfied
        elif self.children:
            return all(c.satisfied for c in self.children)
        else: # pragma: no cover
            return False

    @property
    def satisfied_at(self) -> int:
        """
        ``int``: the step counter value at which this annotation was satisfied; if child results are 
        present, this is the maximum satisfying timestamp of all child results; if this annotation
        was not satisfied, returns -1
        """
        if not self.satisfied:
            return -1
        if self.children:
            return max(c.satisfied_at for c in self.children)
        return self.timestamp

    @property
    def name(self) -> Optional[str]:
        """
        ``str`` or ``None``: the name of the annotation that these results track
        """
        return self.annotation.name
    
    @property
    def group(self) -> Optional[str]:
        """
        ``str`` or ``None``: the group name of the annotation that these results track
        """
        return self.annotation.group

    @property
    def value(self) -> Any:
        """
        ``object``: the value that satisfied the condition of this annotation
        """
        if self._value is None and self.children and len(self.children) == 1:
            return self.children[0].value
        return self._value

    @property
    def messages(self) -> List[Tuple[str, Optional[str], bool]]: # (message, name, satisfied)
        """
        ``list[tuple[str, Optional[str], bool]]``: The messages returned by this annotation and its 
        children based on whether or not the annotations were satisfied. A list of tuples where the 
        first element is the message, the second is the name of the annotation (or ``None`` if no 
        name is  present), and the third is whether the annotation was satisfied.
        """
        messages = []
        for c in self.children:
            messages.extend(c.messages)
        
        if self.satisfied and self.annotation.success_message:
            messages.append((self.annotation.success_message, self.annotation.name, True))
        elif not self.satisfied and self.annotation.failure_message:
            messages.append((self.annotation.failure_message, self.annotation.name, False))

        return messages

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts this annotation result's details to a JSON-friendly dictionary format.

        Output dictionary contains the annotation, whether it was satisfied, when it was satisfied,
        and any child results.

        Returns:
            ``dict[str, object]``: the dictionary representation of this annotation
        """
        return {
            "satisfied": self.satisfied,
            "satisfied_at": self.satisfied_at,
            "annotation": self.annotation.to_dict(),
            "children": [c.to_dict() for c in self.children],
        }


from .relation import *



""""""

import pytest
import numpy as np

from pybryt import check_time_complexity
from pybryt.execution import create_collector, TimeComplexityResult

from .utils import generate_mocked_frame


def test_time_complexity():
    def do_work(n, transform):
        n = int(transform(n))
        for _ in range(n):
            2 * 2
        return n

    (observed, _), cir = create_collector()

    for e in range(1, 9):
        n = 10 ** e
        with check_time_complexity("foo", n):
            do_work(n, lambda v: v)

    assert len(observed) == 8
    assert all(isinstance(t[0], TimeComplexityResult) for t in observed)

    observed.clear()

    for e in range(1, 9):
        n = float(10 ** e)
        with check_time_complexity("foo", n):
            do_work(n, np.log2)

        n = np.random.uniform(size=int(n))
        with check_time_complexity("bar", n):
            do_work(n, len)

    assert len(observed) == 16
    assert all(isinstance(t[0], TimeComplexityResult) for t in observed)
    assert [t[0].name for t in observed] == ["foo", "bar"] * 8

    # test error
    n = lambda v: v
    with pytest.raises(TypeError, match=f"n has invalid type {type(n)}"):
        with check_time_complexity("foo", n):
            pass

    # check that tracking is disabled
    frame = generate_mocked_frame("<ipython-abc-123>", "bar", 1)
    with check_time_complexity("foo", 10):
        cir(frame, "return", 10)
        assert len(observed) == 16
